import importlib.util
from pathlib import Path

from fennil.app.registry import FIELD_REGISTRY

FIELDS_DIRECTORY = Path(__file__).parent / "fields"


def load_all_viz():
    for f in sorted(FIELDS_DIRECTORY.glob("*.py")):
        if f.name.startswith("_"):
            continue

        name = f.stem
        spec = importlib.util.spec_from_file_location(name, f)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)

        FIELD_REGISTRY.register(
            name,
            module.SPEC,
            module.builder,
            module.can_render,
        )


__all__ = [
    "load_all_viz",
]



from fennil.app.io import Dataset
from fennil.app.registry import FieldSpec, LayerContext
from fennil.app.viz.faults import (
    REQUIRED_SEG_COLS,
    fault_line_dataframe,
    segment_slip_layers,
)

SPEC = FieldSpec(
    priority=20,
    label="Slip",
    icon="mdi-chart-line-variant",
    ui_type="VBtnToggle",
    options=[
        {"text": "SS", "value": "ss"},
        {"text": "DS", "value": "ds"},
    ],
    default=None,
    styles={
        "icon_color": "#1976D2",
    },
)


def builder(name: str, ctx: LayerContext):
    for idx, dataset in ctx.enabled_datasets(name):
        folder_number = idx + 1
        seg_tooltip_enabled = True
        fault_lines_df = fault_line_dataframe(dataset.data.segment, seg_tooltip_enabled)
        ctx.layers.extend(
            segment_slip_layers(
                folder_number,
                dataset.data.segment,
                dataset.fields[name],
                seg_tooltip_enabled,
                fault_lines_df,
                ctx.velocity_scale,
            )
        )


def can_render(dataset: Dataset) -> bool:
    return dataset is not None and REQUIRED_SEG_COLS.issubset(dataset.segment.columns)



import pydeck as pdk

from fennil.app.geo_projs import shift_longitudes_df, shift_polygon_df


def line_layers(
    layer_id_prefix,
    data_df,
    get_color,
    line_width,
    folder_number,
    width_min_pixels=1,
    width_max_pixels=None,
    width_scale=1,
    width_units=None,
    pickable=False,
):
    layer_kwargs = {
        "data": data_df,
        "get_source_position": ["start_lon", "start_lat"],
        "get_target_position": ["end_lon", "end_lat"],
        "get_color": get_color,
        "get_width": line_width,
        "width_min_pixels": width_min_pixels,
        "width_scale": width_scale,
        "pickable": pickable,
    }
    if width_max_pixels is not None:
        layer_kwargs["width_max_pixels"] = width_max_pixels
    if width_units is not None:
        # pydeck expects quoted string literals for enum-like values.
        layer_kwargs["width_units"] = f"'{width_units}'"

    layers = [
        pdk.Layer(
            "LineLayer",
            id=f"{layer_id_prefix}_{folder_number}",
            **layer_kwargs,
        )
    ]

    shift_df = shift_longitudes_df(data_df, ["start_lon", "end_lon"])
    shifted_kwargs = layer_kwargs | {"data": shift_df}
    layers.append(
        pdk.Layer(
            "LineLayer",
            id=f"{layer_id_prefix}_shift_{folder_number}",
            **shifted_kwargs,
        )
    )
    return layers


def polygon_layers(
    layer_id_prefix,
    data_df,
    fill_color,
    line_color,
    line_width,
    folder_number,
    line_width_min_pixels=1,
    stroked=True,
    pickable=True,
):
    layers = [
        pdk.Layer(
            "PolygonLayer",
            data=data_df,
            get_polygon="polygon",
            get_fill_color=fill_color,
            get_line_color=line_color,
            get_line_width=line_width,
            filled=True,
            stroked=stroked,
            line_width_min_pixels=line_width_min_pixels,
            pickable=pickable,
            id=f"{layer_id_prefix}_{folder_number}",
        )
    ]

    shift_df = shift_polygon_df(data_df)
    layers.append(
        pdk.Layer(
            "PolygonLayer",
            data=shift_df,
            get_polygon="polygon",
            get_fill_color=fill_color,
            get_line_color=line_color,
            get_line_width=line_width,
            filled=True,
            stroked=stroked,
            line_width_min_pixels=line_width_min_pixels,
            pickable=pickable,
            id=f"{layer_id_prefix}_shift_{folder_number}",
        )
    )
    return layers


def scatter_layers(
    layer_id_prefix,
    data_df,
    fill_color,
    radius,
    folder_number,
    radius_min_pixels=1,
    radius_max_pixels=10,
    pickable=False,
):
    layers = [
        pdk.Layer(
            "ScatterplotLayer",
            data=data_df,
            get_position=["lon", "lat"],
            get_fill_color=fill_color,
            get_radius=radius,
            radius_min_pixels=radius_min_pixels,
            radius_max_pixels=radius_max_pixels,
            pickable=pickable,
            id=f"{layer_id_prefix}_{folder_number}",
        )
    ]

    shift_df = shift_longitudes_df(data_df, ["lon"])
    layers.append(
        pdk.Layer(
            "ScatterplotLayer",
            data=shift_df,
            get_position=["lon", "lat"],
            get_fill_color=fill_color,
            get_radius=radius,
            radius_min_pixels=radius_min_pixels,
            radius_max_pixels=radius_max_pixels,
            pickable=pickable,
            id=f"{layer_id_prefix}_shift_{folder_number}",
        )
    )
    return layers


def icon_layers(
    layer_id_prefix,
    data_df,
    get_position,
    get_icon,
    get_color,
    get_size,
    folder_number,
    position_lon_key="lon",
    get_angle=None,
    size_min_pixels=1,
    size_max_pixels=64,
    billboard=True,
    pickable=False,
):
    layer_kwargs = {
        "data": data_df,
        "get_position": get_position,
        "get_icon": get_icon,
        "get_color": get_color,
        "get_size": get_size,
        "size_min_pixels": size_min_pixels,
        "size_max_pixels": size_max_pixels,
        "billboard": billboard,
        "pickable": pickable,
        "id": f"{layer_id_prefix}_{folder_number}",
    }
    if get_angle is not None:
        layer_kwargs["get_angle"] = get_angle

    layers = [pdk.Layer("IconLayer", **layer_kwargs)]

    shift_df = shift_longitudes_df(data_df, [position_lon_key])
    shifted_kwargs = layer_kwargs | {
        "data": shift_df,
        "id": f"{layer_id_prefix}_shift_{folder_number}",
    }
    layers.append(pdk.Layer("IconLayer", **shifted_kwargs))
    return layers



from .faults import fault_line_layers
from .styles import dataset_float_value, dataset_value, normalize_dataset_values

# Keep base fault colors stable and distinct per dataset.
FAULT_LINES_STYLE_KEY = "fault_lines"
FAULT_LINE_COLORS_DEFAULT = [
    (0, 0, 255, 255),
    (0, 128, 0, 255),
]
FAULT_LINE_WIDTHS_DEFAULT = (1, 1)


def fault_line_style_spec():
    return {
        "priority": 25,
        "label": "Fault lines",
        "icon": "mdi-map-marker-path",
        "type": "VCheckbox",
        "styles": {
            "colors": [list(color) for color in FAULT_LINE_COLORS_DEFAULT],
            "line_width": list(FAULT_LINE_WIDTHS_DEFAULT),
        },
        "options": None,
        "multiple": False,
    }


def build_fault_lines(ctx):
    styles = ctx.specs.get(FAULT_LINES_STYLE_KEY, {}).get("styles", {})
    line_colors = normalize_dataset_values(
        styles.get("colors"), FAULT_LINE_COLORS_DEFAULT
    )
    line_widths = normalize_dataset_values(
        styles.get("line_width"),
        FAULT_LINE_WIDTHS_DEFAULT,
    )

    for idx, dataset in enumerate(ctx.datasets):
        if not dataset.enabled or dataset.data is None:
            continue

        folder_number = idx + 1
        seg_tooltip_enabled = True
        color = dataset_value(line_colors, idx)
        width = dataset_float_value(
            line_widths,
            idx,
            dataset_value(FAULT_LINE_WIDTHS_DEFAULT, idx),
        )
        fault_layers, _ = fault_line_layers(
            folder_number,
            dataset.data.segment,
            seg_tooltip_enabled,
            color,
            width,
        )
        ctx.layers.extend(fault_layers)



from urllib.parse import quote

import numpy as np
import pandas as pd

from fennil.app.deck.primitives import icon_layers, line_layers
from fennil.app.geo_projs import normalize_longitude_difference, web_mercator_to_wgs84

from .styles import (
    VECTOR_ARROW_MAX_PIXELS,
    VECTOR_ARROW_MIN_PIXELS,
    VECTOR_ARROW_SIZE_FACTOR,
    VELOCITY_SCALE,
)

ARROW_ICON = {
    "url": "data:image/svg+xml;utf8,"
    + quote(
        "<svg xmlns='http://www.w3.org/2000/svg' width='64' height='64' viewBox='0 0 64 64'>"
        "<polygon fill='black' points='32,4 60,60 4,60'/>"
        "</svg>"
    ),
    "width": 64,
    "height": 64,
    "anchorY": 64,
    "mask": True,
}


def velocity_layers(
    layer_id_prefix,
    station,
    x_station,
    y_station,
    east_component,
    north_component,
    base_color,
    line_width,
    folder_number,
    velocity_scale,
):
    """Build velocity lines and matching arrowhead tips (including -360 duplicates)."""
    x_station = np.asarray(x_station)
    y_station = np.asarray(y_station)
    east_component = np.asarray(east_component)
    north_component = np.asarray(north_component)

    start_lon = station.lon.to_numpy()
    start_lat = station.lat.to_numpy()

    scale = velocity_scale * VELOCITY_SCALE
    x_end = x_station + scale * east_component
    y_end = y_station + scale * north_component
    end_lon, end_lat = web_mercator_to_wgs84(x_end, y_end)
    end_lon = normalize_longitude_difference(start_lon, end_lon)

    base_df = pd.DataFrame(
        {
            "start_lon": start_lon,
            "start_lat": start_lat,
            "end_lon": end_lon,
            "end_lat": end_lat,
        }
    )
    layers = line_layers(
        layer_id_prefix,
        base_df,
        base_color,
        line_width,
        folder_number,
        width_min_pixels=1,
        pickable=False,
    )

    vector_magnitude = np.hypot(east_component, north_component)
    arrow_mask = np.isfinite(vector_magnitude) & (vector_magnitude > 0)
    if not np.any(arrow_mask):
        return layers

    # IconLayer rotation is counter-clockwise; convert from clockwise bearing.
    angle = -np.degrees(np.arctan2(east_component, north_component)) % 360.0
    arrow_count = int(np.count_nonzero(arrow_mask))

    arrow_df = pd.DataFrame(
        {
            "lon": end_lon[arrow_mask],
            "lat": end_lat[arrow_mask],
            "angle": angle[arrow_mask],
            "icon": [ARROW_ICON] * arrow_count,
        }
    )
    arrow_size = float(
        np.clip(
            line_width * VECTOR_ARROW_SIZE_FACTOR,
            VECTOR_ARROW_MIN_PIXELS,
            VECTOR_ARROW_MAX_PIXELS,
        )
    )
    layers.extend(
        icon_layers(
            f"{layer_id_prefix}_arrow",
            arrow_df,
            get_position=["lon", "lat"],
            get_icon="icon",
            get_color=base_color,
            get_size=arrow_size,
            get_angle="angle",
            folder_number=folder_number,
            position_lon_key="lon",
            size_min_pixels=VECTOR_ARROW_MIN_PIXELS,
            size_max_pixels=VECTOR_ARROW_MAX_PIXELS,
            billboard=False,
            pickable=False,
        )
    )

    return layers



from fennil.app.io import Dataset
from fennil.app.registry import FieldSpec, LayerContext
from fennil.app.viz.vectors import velocity_layers

SPEC = FieldSpec(
    priority=15,
    label="Tri",
    icon="mdi-vector-line",
    ui_type="VCheckbox",
    options=None,
    default=False,
    styles={
        "icon_color": "rgba(205, 133, 0, 0.78)",
        "colors": [
            (205, 133, 0, 200),
            (205, 133, 0, 200),
        ],
        "line_width": (1, 2),
    },
)


def builder(name: str, ctx: LayerContext):
    if ctx.skip(name):
        return

    for idx, dataset in ctx.enabled_datasets(name):
        ctx.vector_layers.extend(
            velocity_layers(
                "tde_vel",
                dataset.data.station,
                dataset.data.x_station,
                dataset.data.y_station,
                dataset.data.station.model_east_vel_tde.values,
                dataset.data.station.model_north_vel_tde.values,
                ctx.specs[name]["styles"]["colors"][idx],
                ctx.specs[name]["styles"]["line_width"][idx],
                dataset.name,
                ctx.velocity_scale,
            )
        )


def can_render(dataset: Dataset) -> bool:
    return dataset is not None



from fennil.app.io import Dataset
from fennil.app.registry import FieldSpec, LayerContext
from fennil.app.viz.res_compare import residual_compare_layers
from fennil.app.viz.styles import RES_COMPARE_DIFF_MAX, RES_COMPARE_DIFF_MIN

SPEC = FieldSpec(
    priority=50,
    label="Res compare",
    icon="mdi-circle-multiple",
    ui_type="VCheckbox",
    options=None,
    default=False,
    styles={
        "icon_color": "rgba(205, 0, 205, 0.78)",
        "color_map_range": {
            "palette": "RDBU_11",
            "min": RES_COMPARE_DIFF_MIN,
            "max": RES_COMPARE_DIFF_MAX,
        },
    },
    multiple=False,
)


def builder(name: str, ctx: LayerContext):
    right = ctx.datasets[0]
    left = ctx.datasets[1]
    if not (right.enabled and left.enabled):
        return
    if right.data is None or left.data is None:
        return
    if not left.fields.get(name):
        return

    styles = ctx.specs[name].get("styles", {})
    color_map_range = styles.get("color_map_range", {})
    try:
        value_min = float(color_map_range.get("min", RES_COMPARE_DIFF_MIN))
        value_max = float(color_map_range.get("max", RES_COMPARE_DIFF_MAX))
    except (TypeError, ValueError):
        value_min = RES_COMPARE_DIFF_MIN
        value_max = RES_COMPARE_DIFF_MAX

    ctx.layers.extend(
        residual_compare_layers(
            right.data,
            left.data,
            ctx.velocity_scale,
            value_min=value_min,
            value_max=value_max,
        )
    )


def can_render(dataset: Dataset) -> bool:
    return dataset is not None



import pydeck as pdk

from . import mapbox


def build_deck(layers, map_params):
    return pdk.Deck(
        map_provider=mapbox.PROVIDER,
        map_style=mapbox.STYLE,
        initial_view_state=pdk.ViewState(
            latitude=map_params.latitude,
            longitude=map_params.longitude,
            zoom=map_params.zoom,
            pitch=map_params.pitch,
            bearing=map_params.bearing,
        ),
        layers=layers,
    )



from trame.app import TrameApp
from trame.decorators import change
from trame.ui.vuetify3 import VAppLayout
from trame.widgets import deckgl, html
from trame.widgets import vuetify3 as v3
from trame_dataclass.core import get_instance

from fennil.app.io import load_folder_data

from .components import FileBrowser, Scale, StyleEditor
from .deck import build_deck, mapbox
from .registry import FIELD_REGISTRY, LayerContext
from .state import DatasetVisualization, MapSettings
from .viz import load_all_viz
from .viz.fault_lines import (
    FAULT_LINES_STYLE_KEY,
    build_fault_lines,
    fault_line_style_spec,
)


class FennilApp(TrameApp):
    def __init__(self, server=None):
        super().__init__(server, client_type="vue3")

        # --hot-reload for dev UI faster
        if self.server.hot_reload:
            self.server.controller.on_server_reload.add(self._build_ui)

        # Load all available viz
        load_all_viz()

        # Only 2 datasets max
        self._datasets = [
            DatasetVisualization(self.server),
            DatasetVisualization(self.server),
        ]
        self.map_params = MapSettings(self.server)
        for viz_config in self._datasets:
            viz_config.watch(["fields", "enabled"], self._update_layers)
        field_specs = FIELD_REGISTRY.export_specs()
        field_specs[FAULT_LINES_STYLE_KEY] = fault_line_style_spec()
        self.state.field_specs = field_specs
        self.style_editor = StyleEditor(self, self._datasets)

        # build ui
        self._build_ui()
        self.style_editor.refresh()

    @change("scale", "field_specs")
    def _update_layers(self, *_, **__):
        """Update DeckGL layers based on loaded data and visibility controls"""
        ctx = LayerContext(
            specs=self.state.field_specs,
            datasets=self._datasets,
            velocity_scale=self.state.scale,
        )
        build_fault_lines(ctx)
        FIELD_REGISTRY.build_layers(ctx)

        with self.state:
            self.ctrl.deck_update(build_deck(ctx.all_layers, self.map_params))

    def load_dataset(self, directory_path):
        self.state.compact_drawer = False  # Always open when new data
        dataset = load_folder_data(directory_path)
        if self._datasets[0].enabled:
            self._datasets[1].attach_data(directory_path, dataset)
        else:
            self._datasets[0].attach_data(directory_path, dataset)
        self.style_editor.refresh()

    def reset_dataset(self, index):
        if index == 0 and self._datasets[1].enabled:
            self._datasets[0].adopt(self._datasets[1])
            self._datasets[1].clear()
            self.style_editor.refresh()
            return
        self._datasets[index].clear()
        self.style_editor.refresh()

    def update_dataset_config(self, id, name, value):
        """Keep server in sync with client reactive nested structure"""
        state = get_instance(id)
        state.fields = {**state.fields, name: value}

    def _build_ui(self, **_):
        self.state.trame__title = "Earthquake Data Viewer"
        with VAppLayout(self.server, fill_height=True) as self.ui:
            # -----------------------------------------------------------------
            # Dialogs
            # -----------------------------------------------------------------

            FileBrowser(ctx_name="file_browser", on_open=self.load_dataset)

            # -----------------------------------------------------------------
            # Drawer
            # -----------------------------------------------------------------

            with v3.VNavigationDrawer(
                permanent=True,
                rail=("compact_drawer", False),
                width=340,
            ):
                with v3.Template(v_slot_prepend=True):
                    with v3.VListItem(
                        title=["compact_drawer ? null : 'Earthquake Viewer'"],
                        click="compact_drawer = !compact_drawer",
                        prepend_icon="mdi-map",
                    ):
                        v3.VProgressCircular(
                            color="primary",
                            indeterminate=True,
                            v_show="trame__busy",
                            v_if="compact_drawer",
                            style="position: absolute !important;left: 50%;top: 50%; transform: translate(-50%, -50%);",
                        )
                        v3.VProgressLinear(
                            v_else=True,
                            color="primary",
                            indeterminate=True,
                            v_show="trame__busy",
                            absolute=True,
                            style="top:90%;width:100%;",
                        )
                    v3.VListItem(
                        title=["compact_drawer ? null : 'Load Dataset'"],
                        click=self.ctx.file_browser.open,
                        prepend_icon="mdi-database-plus",
                    )
                    v3.VListItem(
                        title=["compact_drawer ? null : 'Style Editor'"],
                        click=self.style_editor.open,
                        prepend_icon="mdi-tune-variant",
                    )

                with self._datasets[0].provide_as("right"):
                    with self._datasets[1].provide_as("left"):
                        with v3.VTable(
                            v_if="!compact_drawer && (left.enabled || right.enabled)",
                            density="compact",
                            striped="even",
                            fixed_header=True,
                            height="calc(100vh - 104px - 40px - 40px)",
                        ):
                            with html.Thead():
                                with html.Tr():
                                    html.Th("Field", classes="text-center")
                                    with html.Th():
                                        with html.Div(classes="d-flex align-center"):
                                            v3.VBtn(
                                                icon="mdi-trash-can-outline",
                                                density="compact",
                                                hide_details=True,
                                                size="small",
                                                variant="plain",
                                                click=(
                                                    self.reset_dataset,
                                                    "[left.enabled ? 1 : 0]",
                                                ),
                                            )
                                            html.Div(
                                                "{{ left.enabled ? left.name : right.name }}"
                                            )
                                    with html.Th(v_if="left.enabled"):
                                        with html.Div(classes="d-flex align-center"):
                                            v3.VBtn(
                                                icon="mdi-trash-can-outline",
                                                density="compact",
                                                hide_details=True,
                                                size="small",
                                                variant="plain",
                                                click=(self.reset_dataset, "[0]"),
                                            )
                                            html.Div("{{ right.name }}")

                            table_field_rows_expr = (
                                "(left.enabled ? left.available_fields.concat("
                                "right.available_fields.filter((name) => !left.available_fields.includes(name))"
                                ") : right.available_fields)"
                                ".slice().sort((a, b) => (field_specs[a]?.priority ?? 999) - (field_specs[b]?.priority ?? 999))"
                            )
                            table_columns_expr = "left.enabled ? [{data: left, idx: 1}, {data: right, idx: 0}] : [{data: right, idx: 0}]"

                            with html.Tbody():
                                with html.Tr(
                                    v_for=f"name, i in {table_field_rows_expr}",
                                    key="i",
                                    v_show="field_specs[name].multiple || left.enabled",
                                ):
                                    with html.Td():
                                        with html.Div(
                                            classes="d-flex align-center",
                                            v_if="field_specs[name]",
                                        ):
                                            v3.VIcon(
                                                classes="mr-1",
                                                icon=["field_specs[name]?.icon"],
                                                color=[
                                                    "field_specs[name].styles.icon_color"
                                                ],
                                            )
                                            v3.VLabel(
                                                "{{ field_specs[name].label || name }}",
                                                classes="text-capitalize",
                                            )
                                    with v3.Template(v_if="field_specs[name].multiple"):
                                        with html.Td(
                                            v_for=f"col, i in {table_columns_expr}",
                                            key="i",
                                            v_show="col.data.enabled",
                                        ):
                                            with html.Div(
                                                classes="d-flex align-center justify-center",
                                                v_if="col.data.enabled",
                                            ):
                                                v3.VCheckbox(
                                                    v_if="col.data.available_fields.includes(name) && field_specs[name]?.type === 'VCheckbox'",
                                                    v_model="col.data.fields[name]",
                                                    hide_details=True,
                                                    density="compact",
                                                    update_modelValue=(
                                                        self.update_dataset_config,
                                                        "[col.data._id, name, col.data.fields[name]]",
                                                    ),
                                                )
                                                with v3.VBtnToggle(
                                                    v_if="col.data.available_fields.includes(name) && field_specs[name]?.type === 'VBtnToggle'",
                                                    v_model="col.data.fields[name]",
                                                    hide_details=True,
                                                    density="compact",
                                                    rounded="md",
                                                    border=True,
                                                    divided=True,
                                                    style="height: 24px;",
                                                    update_modelValue=(
                                                        self.update_dataset_config,
                                                        "[col.data._id, name, col.data.fields[name]]",
                                                    ),
                                                ):
                                                    v3.VBtn(
                                                        v_for="props, i in field_specs[name].options",
                                                        key="i",
                                                        size=24,
                                                        density="compact",
                                                        hide_details=True,
                                                        v_bind="props",
                                                    )
                                    with html.Td(v_else=True, colspan=2):
                                        with html.Div(
                                            classes="d-flex align-center justify-center",
                                            v_for="data, i in [left.enabled ? left : right]",
                                            key="i",
                                        ):
                                            v3.VCheckbox(
                                                v_if="field_specs[name]?.type === 'VCheckbox'",
                                                v_model="data.fields[name]",
                                                hide_details=True,
                                                density="compact",
                                                update_modelValue=(
                                                    self.update_dataset_config,
                                                    "[data._id, name, data.fields[name]]",
                                                ),
                                            )
                                            with v3.VBtnToggle(
                                                v_if="field_specs[name]?.type === 'VBtnToggle'",
                                                v_model="data.fields[name]",
                                                hide_details=True,
                                                density="compact",
                                                rounded="md",
                                                border=True,
                                                divided=True,
                                                style="height: 24px;",
                                                update_modelValue=(
                                                    self.update_dataset_config,
                                                    "[data._id, name, data.fields[name]]",
                                                ),
                                            ):
                                                v3.VBtn(
                                                    v_for="props, i in field_specs[name].options",
                                                    key="i",
                                                    size=24,
                                                    density="compact",
                                                    hide_details=True,
                                                    v_bind="props",
                                                )

                        with html.Div(
                            v_if="compact_drawer && (left.enabled || right.enabled)",
                            classes="pa-0 d-flex flex-column align-center",
                        ):
                            with html.Div(
                                v_for="data, i in (left.enabled ? [left, right] : [right])",
                                key="i",
                                classes="pa-0 d-flex flex-column align-center",
                                v_show="data.enabled",
                            ):
                                v3.VChip(
                                    "{{ ((field_specs[name].label || name).replace(' compare', ' cmp')) }} {{ typeof data.fields[name] === 'string' ? data.fields[name].toUpperCase() : null }}",
                                    label=True,
                                    classes="text-capitalize my-1",
                                    v_for="name, i in data.available_fields",
                                    key="i",
                                    v_show="data.fields[name]",
                                    size="x-small",
                                    color=["field_specs[name].styles.icon_color"],
                                )

                with v3.Template(v_slot_append=True):
                    Scale(v_if="!compact_drawer")
                    html.Div(
                        "x {{ scale }}",
                        v_else=True,
                        classes="text-center text-caption",
                    )

            self.style_editor.build_ui()

            # -----------------------------------------------------------------
            # Map
            # -----------------------------------------------------------------

            with v3.VMain():
                deck_map = deckgl.Deck(
                    mapbox_api_key=mapbox.TOKEN,
                    tooltip=(
                        "deckgl_tooltip",
                        {
                            "html": "{tooltip}",
                            "style": {
                                "backgroundColor": "rgba(0, 0, 0, 0.85)",
                                "color": "white",
                                "fontSize": "12px",
                            },
                        },
                    ),
                    style="width: 100%; height: 100%;",
                    classes="fill-height",
                )
                self.ctrl.deck_update = deck_map.update
                self.ctrl.deck_update(build_deck([], self.map_params))

            # -----------------------------------------------------------------
            # Footer
            # -----------------------------------------------------------------

            with v3.VFooter(app=True, height=35):
                # Slip rate colorbar placeholder
                html.Div(
                    "Slip rate (mm/yr): -100 ←→ +100",
                    style="font-size: 0.75rem; color: #666;",
                )
                # Residual magnitude colorbar placeholder
                html.Div(
                    "Resid. mag. (mm/yr): 0 → 5",
                    style="font-size: 0.75rem; color: #666;",
                )
                # Residual diff colorbar placeholder
                html.Div(
                    "Resid. diff. (mm/yr): -5 ←→ +5",
                    style="font-size: 0.75rem; color: #666;",
                )
                if self.server.hot_reload:
                    v3.VSpacer()
                    v3.VBtn(
                        icon="mdi-refresh",
                        click=self.ctrl.on_server_reload,
                        density="compact",
                        classes="rounded",
                        flat=True,
                    )



import numpy as np


def format_number(value, precision=4):
    try:
        if not np.isfinite(value):
            return "n/a"
    except TypeError:
        return "n/a"
    return f"{value:.{precision}f}"


def format_segment_tooltip(name, lon1, lat1, lon2, lat2, ss_rate, ds_rate, ts_rate):
    return (
        f"<b>Name</b>: {name}<br/>"
        f"<b>Start</b>: ({format_number(lon1)}, {format_number(lat1)})<br/>"
        f"<b>End</b>: ({format_number(lon2)}, {format_number(lat2)})<br/>"
        f"<b>Strike-Slip Rate</b>: {format_number(ss_rate)}<br/>"
        f"<b>Dip-Slip Rate</b>: {format_number(ds_rate)}<br/>"
        f"<b>Tensile-Slip Rate</b>: {format_number(ts_rate)}"
    )



from dataclasses import dataclass
from pathlib import Path

import numpy as np
import pandas as pd

from fennil.app.geo_projs import (
    DIP_EPS,
    KM2M,
    RADIUS_EARTH,
    VERTICAL_DIP_DEG,
    calculate_fault_bottom_edge,
    cart2sph,
    wgs84_to_web_mercator,
    wrap2360,
)

PROJ_MESH_DIP_THRESHOLD_DEG = 75.0


@dataclass
class Dataset:
    station: pd.DataFrame
    segment: pd.DataFrame
    meshes: pd.DataFrame
    resmag: np.ndarray
    x_station: np.ndarray
    y_station: np.ndarray
    x1_seg: np.ndarray
    y1_seg: np.ndarray
    x2_seg: np.ndarray
    y2_seg: np.ndarray
    tde_available: bool
    tde_df: pd.DataFrame | None
    tde_perim_df: pd.DataFrame | None
    fault_proj_available: bool
    fault_proj_df: pd.DataFrame | None


def is_valid_data_folder(folder_path):
    required = {"model_station.csv", "model_segment.csv", "model_meshes.csv"}
    return all((folder_path / name).is_file() for name in required)


def build_fault_proj_data(segment):
    fault_proj_required = {
        "lon1",
        "lat1",
        "lon2",
        "lat2",
        "dip",
        "locking_depth",
    }
    fault_proj_available = fault_proj_required.issubset(segment.columns)
    if not fault_proj_available:
        return False, None

    fault_proj_polygons = []
    fault_proj_dips = []
    fault_proj_names = []
    for i in range(len(segment)):
        dip_deg = segment["dip"].iloc[i]
        locking_depth = segment["locking_depth"].iloc[i]
        if not np.isfinite(dip_deg) or not np.isfinite(locking_depth):
            continue
        if abs(dip_deg - VERTICAL_DIP_DEG) <= DIP_EPS:
            continue

        lon1 = segment["lon1"].iloc[i]
        lat1 = segment["lat1"].iloc[i]
        lon2 = segment["lon2"].iloc[i]
        lat2 = segment["lat2"].iloc[i]

        lon1_bot, lat1_bot, lon2_bot, lat2_bot = calculate_fault_bottom_edge(
            lon1,
            lat1,
            lon2,
            lat2,
            locking_depth,
            dip_deg,
        )
        fault_proj_polygons.append(
            [
                [lon1, lat1],
                [lon2, lat2],
                [lon2_bot, lat2_bot],
                [lon1_bot, lat1_bot],
                [lon1, lat1],
            ]
        )
        fault_proj_dips.append(float(dip_deg))
        if "name" in segment.columns:
            fault_proj_names.append(str(segment["name"].iloc[i]))
        else:
            fault_proj_names.append("")

    if not fault_proj_polygons:
        return True, None

    fault_proj_df = pd.DataFrame(
        {
            "polygon": fault_proj_polygons,
            "dip": fault_proj_dips,
            "name": fault_proj_names,
        }
    )
    return True, fault_proj_df


def build_tde_data(meshes):
    tde_required = {
        "lon1",
        "lat1",
        "dep1",
        "lon2",
        "lat2",
        "dep2",
        "lon3",
        "lat3",
        "dep3",
        "mesh_idx",
        "strike_slip_rate",
        "dip_slip_rate",
    }
    tde_available = tde_required.issubset(meshes.columns)
    if not tde_available:
        return False, None, None

    lon1_mesh = meshes["lon1"].to_numpy().copy()
    lat1_mesh = meshes["lat1"].to_numpy()
    dep1_mesh = meshes["dep1"].to_numpy()
    lon2_mesh = meshes["lon2"].to_numpy().copy()
    lat2_mesh = meshes["lat2"].to_numpy()
    dep2_mesh = meshes["dep2"].to_numpy()
    lon3_mesh = meshes["lon3"].to_numpy().copy()
    lat3_mesh = meshes["lat3"].to_numpy()
    dep3_mesh = meshes["dep3"].to_numpy()
    mesh_idx = meshes["mesh_idx"].to_numpy()

    lon1_mesh[lon1_mesh < 0] += 360
    lon2_mesh[lon2_mesh < 0] += 360
    lon3_mesh[lon3_mesh < 0] += 360

    tri_leg1 = np.transpose(
        [
            np.deg2rad(lon2_mesh - lon1_mesh),
            np.deg2rad(lat2_mesh - lat1_mesh),
            (1 + KM2M * dep2_mesh / RADIUS_EARTH)
            - (1 + KM2M * dep1_mesh / RADIUS_EARTH),
        ]
    )
    tri_leg2 = np.transpose(
        [
            np.deg2rad(lon3_mesh - lon1_mesh),
            np.deg2rad(lat3_mesh - lat1_mesh),
            (1 + KM2M * dep3_mesh / RADIUS_EARTH)
            - (1 + KM2M * dep1_mesh / RADIUS_EARTH),
        ]
    )
    norm_vec = np.cross(tri_leg1, tri_leg2)
    tri_area = np.linalg.norm(norm_vec, axis=1)
    azimuth, elevation, _r = cart2sph(norm_vec[:, 0], norm_vec[:, 1], norm_vec[:, 2])
    strike = wrap2360(-np.rad2deg(azimuth))
    dip = 90 - np.rad2deg(elevation)
    dip[dip > 90] = 180.0 - dip[dip > 90]

    mesh_list = np.unique(mesh_idx)
    proj_mesh_flag = np.zeros_like(mesh_list)
    mesh_area = np.zeros_like(mesh_list, dtype=float)
    for i in mesh_list:
        this_mesh_els = mesh_idx == i
        mesh_area[i] = np.mean(tri_area[this_mesh_els])
        this_mesh_dip = np.mean(dip[this_mesh_els])
        if this_mesh_dip > PROJ_MESH_DIP_THRESHOLD_DEG:
            proj_mesh_flag[i] = 1
            dip_dir = np.mean(np.deg2rad(strike[this_mesh_els] + 90))
            lon1_mesh[this_mesh_els] += np.sin(dip_dir) * np.rad2deg(
                np.abs(KM2M * dep1_mesh[this_mesh_els] / RADIUS_EARTH)
            )
            lat1_mesh[this_mesh_els] += np.cos(dip_dir) * np.rad2deg(
                np.abs(KM2M * dep1_mesh[this_mesh_els] / RADIUS_EARTH)
            )
            lon2_mesh[this_mesh_els] += np.sin(dip_dir) * np.rad2deg(
                np.abs(KM2M * dep2_mesh[this_mesh_els] / RADIUS_EARTH)
            )
            lat2_mesh[this_mesh_els] += np.cos(dip_dir) * np.rad2deg(
                np.abs(KM2M * dep2_mesh[this_mesh_els] / RADIUS_EARTH)
            )
            lon3_mesh[this_mesh_els] += np.sin(dip_dir) * np.rad2deg(
                np.abs(KM2M * dep3_mesh[this_mesh_els] / RADIUS_EARTH)
            )
            lat3_mesh[this_mesh_els] += np.cos(dip_dir) * np.rad2deg(
                np.abs(KM2M * dep3_mesh[this_mesh_els] / RADIUS_EARTH)
            )

    proj_mesh_idx = np.where(proj_mesh_flag)[0]

    edge1_lon = np.array((lon1_mesh, lon2_mesh))
    edge1_lat = np.array((lat1_mesh, lat2_mesh))
    edge1_array = np.vstack(
        (np.sort(edge1_lon, axis=0), np.sort(edge1_lat, axis=0), mesh_idx)
    )
    edge2_lon = np.array((lon2_mesh, lon3_mesh))
    edge2_lat = np.array((lat2_mesh, lat3_mesh))
    edge2_array = np.vstack(
        (np.sort(edge2_lon, axis=0), np.sort(edge2_lat, axis=0), mesh_idx)
    )
    edge3_lon = np.array((lon3_mesh, lon1_mesh))
    edge3_lat = np.array((lat3_mesh, lat1_mesh))
    edge3_array = np.vstack(
        (np.sort(edge3_lon, axis=0), np.sort(edge3_lat, axis=0), mesh_idx)
    )
    all_edge_array = np.concatenate((edge1_array, edge2_array, edge3_array), axis=1)

    edge1_array_unsorted = np.vstack((edge1_lon, edge1_lat, mesh_idx))
    edge2_array_unsorted = np.vstack((edge2_lon, edge2_lat, mesh_idx))
    edge3_array_unsorted = np.vstack((edge3_lon, edge3_lat, mesh_idx))
    all_edge_array_unsorted = np.concatenate(
        (edge1_array_unsorted, edge2_array_unsorted, edge3_array_unsorted), axis=1
    )

    _unique_edges, unique_edge_index, edge_count = np.unique(
        all_edge_array, return_index=True, return_counts=True, axis=1
    )
    unique_edges_unsorted = all_edge_array_unsorted[:, unique_edge_index]
    perim_edges = unique_edges_unsorted[:, edge_count == 1]
    proj_mesh_edge_flag = np.isin(perim_edges[-1, :], proj_mesh_idx).astype(int)

    mesh_plot_order = np.argsort(-mesh_area)
    mesh_plot_order_index = []
    for i in mesh_plot_order:
        mesh_plot_order_index.extend(np.argwhere(mesh_idx == i).flatten().tolist())
    mesh_plot_order_index = np.array(mesh_plot_order_index, dtype=int)

    tde_df = None
    if mesh_plot_order_index.size:
        tde_data = {
            "polygon": [
                [
                    [lon1_mesh[j], lat1_mesh[j]],
                    [lon2_mesh[j], lat2_mesh[j]],
                    [lon3_mesh[j], lat3_mesh[j]],
                ]
                for j in mesh_plot_order_index
            ],
            "ss_rate": meshes["strike_slip_rate"].to_numpy()[mesh_plot_order_index],
            "ds_rate": meshes["dip_slip_rate"].to_numpy()[mesh_plot_order_index],
        }
        if "strike_slip_coupling" in meshes.columns:
            tde_data["ss_coupling"] = meshes["strike_slip_coupling"].to_numpy()[
                mesh_plot_order_index
            ]
        if "dip_slip_coupling" in meshes.columns:
            tde_data["ds_coupling"] = meshes["dip_slip_coupling"].to_numpy()[
                mesh_plot_order_index
            ]

        tde_df = pd.DataFrame(tde_data)

    tde_perim_df = None
    if perim_edges.size:
        tde_perim_df = pd.DataFrame(
            {
                "start_lon": perim_edges[0, :],
                "start_lat": perim_edges[2, :],
                "end_lon": perim_edges[1, :],
                "end_lat": perim_edges[3, :],
                "proj_col": proj_mesh_edge_flag,
            }
        )

    return True, tde_df, tde_perim_df


def load_folder_data(folder_path):
    folder_path = Path(folder_path)

    station = pd.read_csv(folder_path / "model_station.csv")
    segment = pd.read_csv(folder_path / "model_segment.csv")
    meshes = pd.read_csv(folder_path / "model_meshes.csv")

    resmag = np.sqrt(
        np.power(station.model_east_vel_residual, 2)
        + np.power(station.model_north_vel_residual, 2)
    )

    lon_station = station.lon.to_numpy()
    lat_station = station.lat.to_numpy()
    x_station, y_station = wgs84_to_web_mercator(lon_station, lat_station)

    lon1_seg = segment.lon1.to_numpy()
    lat1_seg = segment.lat1.to_numpy()
    lon2_seg = segment.lon2.to_numpy()
    lat2_seg = segment.lat2.to_numpy()
    x1_seg, y1_seg = wgs84_to_web_mercator(lon1_seg, lat1_seg)
    x2_seg, y2_seg = wgs84_to_web_mercator(lon2_seg, lat2_seg)

    fault_proj_available, fault_proj_df = build_fault_proj_data(segment)
    tde_available, tde_df, tde_perim_df = build_tde_data(meshes)

    return Dataset(
        station=station,
        segment=segment,
        meshes=meshes,
        resmag=resmag,
        x_station=x_station,
        y_station=y_station,
        x1_seg=x1_seg,
        y1_seg=y1_seg,
        x2_seg=x2_seg,
        y2_seg=y2_seg,
        tde_available=tde_available,
        tde_df=tde_df,
        tde_perim_df=tde_perim_df,
        fault_proj_available=fault_proj_available,
        fault_proj_df=fault_proj_df,
    )



import os

import pydeck as pdk
from dotenv import load_dotenv

TOKEN_ENV_KEY = "FENNIL_MAP_BOX_TOKEN"


load_dotenv()
TOKEN = os.getenv(TOKEN_ENV_KEY, "").strip().strip('"').strip("'")
HAS_MAPBOX_TOKEN = bool(TOKEN)

PROVIDER = "mapbox" if HAS_MAPBOX_TOKEN else "carto"
STYLE = "mapbox://styles/mapbox/light-v9" if HAS_MAPBOX_TOKEN else pdk.map_styles.LIGHT

__all__ = [
    "PROVIDER",
    "STYLE",
    "TOKEN",
]



import numpy as np

SHIFT_LON = -360.0
KM2M = 1.0e3
RADIUS_EARTH = 6371000
WEB_MERCATOR_RADIUS = 6378137.0
VERTICAL_DIP_DEG = 90.0
DIP_EPS = 1.0e-6


def wgs84_to_web_mercator(lon, lat):
    """Converts decimal (longitude, latitude) to Web Mercator (x, y)."""
    x = WEB_MERCATOR_RADIUS * np.deg2rad(lon)
    y = WEB_MERCATOR_RADIUS * np.log(np.tan(np.pi / 4.0 + np.deg2rad(lat) / 2.0))
    return x, y


def web_mercator_to_wgs84(x, y):
    """Converts Web Mercator (x, y) to WGS84 (longitude, latitude)."""
    lon = np.rad2deg(x / WEB_MERCATOR_RADIUS)
    lat = np.rad2deg(2.0 * np.arctan(np.exp(y / WEB_MERCATOR_RADIUS)) - np.pi / 2.0)
    return lon, lat


def normalize_longitude_difference(start_lon, end_lon):
    """
    Normalize end longitude to be in the same 360-degree range as start longitude.
    This prevents vectors from wrapping around the world when crossing the date line.
    """
    diff = end_lon - start_lon
    diff = np.where(diff > 180, diff - 360, diff)
    diff = np.where(diff < -180, diff + 360, diff)
    return start_lon + diff


def wrap2360(lon):
    """Wrap longitude to 0-360 range."""
    lon[np.where(lon < 0.0)] += 360.0
    return lon


def calculate_fault_bottom_edge(lon1, lat1, lon2, lat2, depth_km, dip_degrees):
    """Calculate bottom edge coordinates for a dipping fault plane."""
    dip_rad = np.radians(dip_degrees)
    lat1_rad = np.radians(lat1)
    lat2_rad = np.radians(lat2)
    lon1_rad = np.radians(lon1)
    lon2_rad = np.radians(lon2)

    earth_radius_km = RADIUS_EARTH / KM2M

    if np.abs(dip_degrees - VERTICAL_DIP_DEG) < DIP_EPS:
        return lon1, lat1, lon2, lat2

    delta_lon = lon2_rad - lon1_rad
    y = np.sin(delta_lon) * np.cos(lat2_rad)
    x = np.cos(lat1_rad) * np.sin(lat2_rad) - np.sin(lat1_rad) * np.cos(
        lat2_rad
    ) * np.cos(delta_lon)
    strike_bearing = np.arctan2(y, x)

    dip_direction = strike_bearing + np.pi / 2

    min_dip_rad = np.deg2rad(0.1)
    if np.abs(dip_rad) < min_dip_rad:
        return lon1, lat1, lon2, lat2

    horizontal_distance_km = depth_km / np.tan(dip_rad)
    angular_distance = horizontal_distance_km / earth_radius_km

    lat1_bottom_rad = np.arcsin(
        np.sin(lat1_rad) * np.cos(angular_distance)
        + np.cos(lat1_rad) * np.sin(angular_distance) * np.cos(dip_direction)
    )
    lon1_bottom_rad = lon1_rad + np.arctan2(
        np.sin(dip_direction) * np.sin(angular_distance) * np.cos(lat1_rad),
        np.cos(angular_distance) - np.sin(lat1_rad) * np.sin(lat1_bottom_rad),
    )

    lat2_bottom_rad = np.arcsin(
        np.sin(lat2_rad) * np.cos(angular_distance)
        + np.cos(lat2_rad) * np.sin(angular_distance) * np.cos(dip_direction)
    )
    lon2_bottom_rad = lon2_rad + np.arctan2(
        np.sin(dip_direction) * np.sin(angular_distance) * np.cos(lat2_rad),
        np.cos(angular_distance) - np.sin(lat2_rad) * np.sin(lat2_bottom_rad),
    )

    lon1_bottom = np.degrees(lon1_bottom_rad)
    lat1_bottom = np.degrees(lat1_bottom_rad)
    lon2_bottom = np.degrees(lon2_bottom_rad)
    lat2_bottom = np.degrees(lat2_bottom_rad)

    return lon1_bottom, lat1_bottom, lon2_bottom, lat2_bottom


def sph2cart(lon, lat, radius):
    """Convert spherical coordinates to Cartesian."""
    lon_rad = np.deg2rad(lon)
    lat_rad = np.deg2rad(lat)
    x = radius * np.cos(lat_rad) * np.cos(lon_rad)
    y = radius * np.cos(lat_rad) * np.sin(lon_rad)
    z = radius * np.sin(lat_rad)
    return x, y, z


def cart2sph(x, y, z):
    """Convert Cartesian coordinates to spherical."""
    lon = np.arctan2(y, x)
    hyp = np.sqrt(x**2 + y**2)
    lat = np.arctan2(z, hyp)
    r = np.sqrt(x**2 + y**2 + z**2)
    return lon, lat, r


def shift_longitudes_df(data_df, columns, shift=SHIFT_LON):
    shifted = data_df.copy()
    for column in columns:
        shifted[column] = shifted[column] + shift
    return shifted


def shift_polygon_df(data_df, polygon_key="polygon", shift=SHIFT_LON):
    shifted = data_df.copy()
    shifted[polygon_key] = [
        [[pt[0] + shift, pt[1]] for pt in polygon] for polygon in data_df[polygon_key]
    ]
    return shifted



from trame.widgets import vuetify3 as v3

v3.enable_lab()


class Scale(v3.VCol):
    def __init__(self, name="scale", **kwargs):
        self._name = name
        super().__init__(**kwargs)

        with self:
            with v3.VRow(dense=True, no_gutter=True):
                with v3.VCol():
                    v3.VBtn(
                        "50%",
                        click=(self.scale, "[0.5]"),
                        size="small",
                        variant="outlined",
                        block=True,
                    )
                with v3.VCol():
                    v3.VBtn(
                        "90%",
                        click=(self.scale, "[0.9]"),
                        size="small",
                        variant="outlined",
                        block=True,
                    )
                with v3.VCol():
                    v3.VBtn(
                        "110%",
                        click=(self.scale, "[1.1]"),
                        size="small",
                        variant="outlined",
                        block=True,
                    )
                with v3.VCol():
                    v3.VBtn(
                        "200%",
                        click=(self.scale, "[2]"),
                        size="small",
                        variant="outlined",
                        block=True,
                    )
            with v3.VRow():
                with v3.VCol(classes="pt-0"):
                    v3.VNumberInput(
                        click_prepend=self.reset,
                        prepend_icon="mdi-vector-line",
                        v_model=(name, 1),
                        precision=6,
                        min=[0.0001],
                        max=[10000],
                        step=[0.01],
                        density="compact",
                        hide_details=True,
                        variant="outlined",
                        control_variant="split",
                    )

    def scale(self, value):
        self.state[self._name] *= value

    def reset(self):
        self.state[self._name] = 1



from fennil.app.io import Dataset
from fennil.app.registry import FieldSpec, LayerContext
from fennil.app.viz.styles import COUPLING_MAX, COUPLING_MIN
from fennil.app.viz.tde import tde_coupling_layers, tde_perimeter_layers

REQUIRED_COUPLING_COLS = {"strike_slip_coupling", "dip_slip_coupling"}

SPEC = FieldSpec(
    priority=22,
    label="Coupling",
    icon="mdi-link-variant",
    ui_type="VBtnToggle",
    options=[
        {"text": "SS", "value": "ss"},
        {"text": "DS", "value": "ds"},
    ],
    default=None,
    styles={
        "icon_color": "rgba(111, 66, 193, 1)",
        "color_map_range": {
            "palette": "RDBU_11",
            "min": COUPLING_MIN,
            "max": COUPLING_MAX,
        },
    },
)


def builder(name: str, ctx: LayerContext):
    if ctx.skip(name):
        return

    styles = ctx.specs[name].get("styles", {})
    color_map_range = styles.get("color_map_range", {})
    try:
        value_min = float(color_map_range.get("min", COUPLING_MIN))
        value_max = float(color_map_range.get("max", COUPLING_MAX))
    except (TypeError, ValueError):
        value_min = COUPLING_MIN
        value_max = COUPLING_MAX

    for idx, dataset in ctx.enabled_datasets(name):
        folder_number = idx + 1
        value = dataset.fields[name]
        tde_df = dataset.data.tde_df
        tde_perim_df = dataset.data.tde_perim_df
        if tde_df is not None and not tde_df.empty:
            if value == "ss" and "ss_coupling" in tde_df.columns:
                coupling_values = tde_df["ss_coupling"].to_numpy()
            elif value == "ds" and "ds_coupling" in tde_df.columns:
                coupling_values = tde_df["ds_coupling"].to_numpy()
            else:
                coupling_values = None

            if coupling_values is not None:
                ctx.tde_layers.extend(
                    tde_coupling_layers(
                        folder_number,
                        tde_df,
                        coupling_values,
                        value_min,
                        value_max,
                    )
                )
                ctx.tde_layers.extend(tde_perimeter_layers(folder_number, tde_perim_df))


def can_render(dataset: Dataset) -> bool:
    if dataset is None or not dataset.tde_available:
        return False
    return REQUIRED_COUPLING_COLS.issubset(dataset.meshes.columns)



from fennil.app.io import Dataset
from fennil.app.registry import FieldSpec, LayerContext
from fennil.app.viz.faults import fault_projection_layers
from fennil.app.viz.styles import dataset_value, normalize_dataset_values

FAULT_PROJ_FILL_DEFAULT = [
    (173, 216, 230, 77),
    (144, 238, 144, 77),
]
FAULT_PROJ_LINE_DEFAULT = [
    (0, 0, 255, 255),
    (0, 128, 0, 255),
]

SPEC = FieldSpec(
    priority=30,
    label="Fault Proj",
    icon="mdi-bandage",
    ui_type="VCheckbox",
    options=None,
    default=False,
    styles={
        "fill": FAULT_PROJ_FILL_DEFAULT,
    },
)


def builder(name: str, ctx: LayerContext):
    if ctx.skip(name):
        return

    styles = ctx.specs[name].get("styles", {})
    fill_values = normalize_dataset_values(styles.get("fill"), FAULT_PROJ_FILL_DEFAULT)
    line_values = normalize_dataset_values(styles.get("line"), FAULT_PROJ_LINE_DEFAULT)

    for idx, dataset in ctx.enabled_datasets(name):
        ctx.layers.extend(
            fault_projection_layers(
                dataset.name,
                dataset.data.fault_proj_df,
                dataset_value(fill_values, idx),
                dataset_value(line_values, idx),
            )
        )


def can_render(dataset: Dataset) -> bool:
    if dataset is None:
        return False

    return dataset.fault_proj_available



from pathlib import Path

from trame_dataclass.core import StateDataModel

from fennil.app.registry import FIELD_REGISTRY

DEFAULT_VIEW_STATE = {
    "latitude": 35.0,
    "longitude": -165.0,
    "zoom": 2,
    "pitch": 0,
    "bearing": 0,
}


class MapSettings(StateDataModel):
    latitude: float = DEFAULT_VIEW_STATE["latitude"]
    longitude: float = DEFAULT_VIEW_STATE["longitude"]
    zoom: float = DEFAULT_VIEW_STATE["zoom"]
    pitch: float = DEFAULT_VIEW_STATE["pitch"]
    bearing: float = DEFAULT_VIEW_STATE["bearing"]


class DatasetVisualization(StateDataModel):
    enabled: bool = False
    name: str
    fields: dict[str, bool | str | None]
    available_fields: list[str]

    def attach_data(self, directory_path, data):
        self._data = data
        self.name = ""
        self.enabled = bool(data)
        if data:
            self.name = Path(directory_path).stem.lstrip("0")
            self.available_fields = FIELD_REGISTRY.available_fields(data)
            self.fields = FIELD_REGISTRY.field_defaults()

    @property
    def data(self):
        return getattr(self, "_data", None)

    def clear(self):
        self.enabled = False
        self.fields = {}
        self.available_fields = []

    def adopt(self, other):
        self._data = other.data
        self.name = other.name
        self.enabled = other.enabled
        self.fields = dict(other.fields)
        self.available_fields = list(other.available_fields)



from fennil.app.io import Dataset
from fennil.app.registry import FieldSpec, LayerContext
from fennil.app.viz.styles import SLIP_RATE_MAX, SLIP_RATE_MIN
from fennil.app.viz.tde import tde_mesh_layers, tde_perimeter_layers

SPEC = FieldSpec(
    priority=21,
    label="TDE",
    icon="mdi-texture-box",
    ui_type="VBtnToggle",
    options=[
        {"text": "SS", "value": "ss"},
        {"text": "DS", "value": "ds"},
    ],
    default=None,
    styles={
        "icon_color": "rgba(14, 0, 214, 1)",
        "color_map_range": {
            "palette": "RDBU_11",
            "min": SLIP_RATE_MIN,
            "max": SLIP_RATE_MAX,
        },
    },
)


def builder(name: str, ctx: LayerContext):
    if ctx.skip(name):
        return

    styles = ctx.specs[name].get("styles", {})
    color_map_range = styles.get("color_map_range", {})
    try:
        value_min = float(color_map_range.get("min", SLIP_RATE_MIN))
        value_max = float(color_map_range.get("max", SLIP_RATE_MAX))
    except (TypeError, ValueError):
        value_min = SLIP_RATE_MIN
        value_max = SLIP_RATE_MAX

    for idx, dataset in ctx.enabled_datasets(name):
        folder_number = idx + 1
        value = dataset.fields[name]
        tde_df = dataset.data.tde_df
        tde_perim_df = dataset.data.tde_perim_df
        if tde_df is not None and not tde_df.empty:
            slip_values = (
                tde_df["ss_rate"].to_numpy()
                if value == "ss"
                else tde_df["ds_rate"].to_numpy()
            )
            ctx.tde_layers.extend(
                tde_mesh_layers(
                    folder_number,
                    tde_df,
                    slip_values,
                    value_min,
                    value_max,
                )
            )
        ctx.tde_layers.extend(tde_perimeter_layers(folder_number, tde_perim_df))


def can_render(dataset: Dataset) -> bool:
    return dataset is not None and dataset.tde_available



from fennil.app.io import Dataset
from fennil.app.registry import FieldSpec, LayerContext
from fennil.app.viz.stations import station_layers

SPEC = FieldSpec(
    priority=0,
    label="Locs",
    icon="mdi-circle-medium",
    ui_type="VCheckbox",
    options=None,
    default=False,
    styles={
        "icon_color": "black",
        "colors": [
            (0, 0, 0, 220),
            (0, 0, 0, 220),
        ],
        "line_width": (1, 2),
    },
)


def builder(name: str, ctx: LayerContext):
    if ctx.skip(name):
        return

    for idx, dataset in ctx.enabled_datasets(name):
        ctx.layers.extend(
            station_layers(
                dataset.name,
                dataset.data.station,
                ctx.specs[name]["styles"]["colors"][idx],
            )
        )


def can_render(dataset: Dataset) -> bool:
    return dataset is not None



import copy

from trame.widgets import html
from trame.widgets import vuetify3 as v3

from fennil.app.viz.styles import RDBU_11, normalize_dataset_values

PALETTES = {
    "RDBU_11": RDBU_11,
}
DEFAULT_PICKER_HEXA = "#000000ff"
PICKER_STATE_KEYS = (
    "style_editor_picker_open",
    "style_editor_picker_hexa",
    "style_editor_picker_field_idx",
    "style_editor_picker_group_idx",
    "style_editor_picker_entry_idx",
)
COLOR_STYLE_GROUPS = (
    ("colors", "Color"),
    ("fill", "Fill"),
    ("line", "Line"),
)
DATASET_GRID_STYLE = "`display:grid;grid-template-columns:repeat(${group.values.length}, minmax(0, 1fr));`"
DATASET_CARD_STYLE = "border: 1px solid rgba(127,127,127,0.35); border-radius: 8px;"
COLOR_SWATCH_STYLE = "`width:32px;height:32px;min-width:32px;border-radius:50%;border:1px solid #666;background:${entry.hexa};`"
COLORMAP_BAR_STYLE = "`height: 14px; border: 1px solid #666; border-radius: 4px; background: ${group.gradient};`"
MINMAX_GRID_STYLE = "display:grid;grid-template-columns:minmax(0, 1fr) minmax(0, 1fr);"


class StyleEditor:
    def __init__(self, app, datasets):
        self._state = app.state
        self._datasets = datasets
        self._init_state()

    def _init_state(self):
        # Keep large editable form state client-side to avoid server echo jitter
        # while users type in text fields.
        for key in ("style_editor_controls", *PICKER_STATE_KEYS):
            self._state.client_only(key)
        self._state.show_style_editor = False
        self._state.style_editor_controls = []
        self._state.style_editor_picker_open = False
        self._state.style_editor_picker_hexa = DEFAULT_PICKER_HEXA
        self._state.style_editor_picker_field_idx = -1
        self._state.style_editor_picker_group_idx = -1
        self._state.style_editor_picker_entry_idx = -1
        self._state.style_editor_error = ""

    def build_ui(self):
        with v3.VNavigationDrawer(
            location="right",
            temporary=True,
            width=560,
            v_model=("show_style_editor", False),
        ):
            with v3.VToolbar(density="compact", flat=True):
                v3.VToolbarTitle("Style Editor")
                v3.VSpacer()
                v3.VBtn(
                    icon="mdi-close",
                    variant="text",
                    click="show_style_editor = false",
                )

            with v3.VContainer(classes="pa-3"):
                with v3.VExpansionPanels(multiple=True):
                    with v3.VExpansionPanel(
                        v_for="(field, field_idx) in style_editor_controls",
                        key="field.name",
                    ):
                        with v3.VExpansionPanelTitle():
                            with html.Div(classes="d-flex align-center ga-2"):
                                v3.VIcon(
                                    icon=["field.icon"],
                                    color=["field.icon_color || undefined"],
                                    size="small",
                                )
                                html.Div("{{ field.label }}", classes="text-subtitle-2")
                        with v3.VExpansionPanelText():
                            with html.Div(
                                v_for="(group, group_idx) in field.color_groups",
                                key="group.key",
                                classes="mb-2",
                            ):
                                html.Div(
                                    "{{ group.label }}", classes="text-caption mb-1"
                                )
                                with html.Div(
                                    classes="ga-3",
                                    style=[DATASET_GRID_STYLE],
                                ):
                                    with html.Div(
                                        v_for="(entry, entry_idx) in group.values",
                                        key="entry.idx",
                                        classes="pa-2",
                                        style=DATASET_CARD_STYLE,
                                    ):
                                        html.Div(
                                            "{{ entry.label }}",
                                            classes="text-caption mb-2",
                                        )
                                        with html.Div(
                                            classes="d-flex align-center ga-2"
                                        ):
                                            v3.VBtn(
                                                "",
                                                variant="flat",
                                                density="compact",
                                                size="small",
                                                style=[COLOR_SWATCH_STYLE],
                                                click=self._open_picker_js(),
                                            )
                                            v3.VTextField(
                                                v_model="entry.hexa",
                                                label="HEXA",
                                                density="compact",
                                                hide_details=True,
                                                variant="outlined",
                                                style="min-width: 0; flex: 1 1 auto;",
                                            )
                            with html.Div(
                                v_for="group in field.width_groups",
                                key="group.key",
                                classes="mb-2",
                            ):
                                html.Div(
                                    "{{ group.label }}", classes="text-caption mb-1"
                                )
                                with html.Div(
                                    classes="ga-3",
                                    style=[DATASET_GRID_STYLE],
                                ):
                                    with html.Div(
                                        v_for="entry in group.values",
                                        key="entry.idx",
                                        classes="pa-2",
                                        style=DATASET_CARD_STYLE,
                                    ):
                                        html.Div(
                                            "{{ entry.label }}",
                                            classes="text-caption mb-2",
                                        )
                                        v3.VNumberInput(
                                            v_model="entry.value",
                                            label="Line Width",
                                            density="compact",
                                            hide_details=True,
                                            variant="outlined",
                                            control_variant="split",
                                            style="min-width: 0;",
                                        )
                            with html.Div(
                                v_for="group in field.colormap_groups",
                                key="group.key",
                                classes="mb-2",
                            ):
                                html.Div(
                                    "{{ group.label }}", classes="text-caption mb-1"
                                )
                                html.Div(
                                    classes="mb-2",
                                    style=[COLORMAP_BAR_STYLE],
                                )
                                with html.Div(
                                    classes="ga-3",
                                    style=MINMAX_GRID_STYLE,
                                ):
                                    v3.VNumberInput(
                                        v_model="group.min",
                                        label="Min",
                                        density="compact",
                                        hide_details=True,
                                        variant="outlined",
                                        control_variant="split",
                                        style="min-width: 0;",
                                    )
                                    v3.VNumberInput(
                                        v_model="group.max",
                                        label="Max",
                                        density="compact",
                                        hide_details=True,
                                        variant="outlined",
                                        control_variant="split",
                                        style="min-width: 0;",
                                    )
                            html.Div(
                                "No editable styles.",
                                v_if="!field.color_groups.length && !field.width_groups.length && !field.colormap_groups.length",
                                classes="text-caption text-medium-emphasis",
                            )
                with v3.VDialog(
                    v_model=("style_editor_picker_open", False),
                    max_width=360,
                ):
                    with v3.VCard():
                        with v3.VCardText(classes="pa-2"):
                            v3.VColorPicker(
                                model_value=(
                                    "style_editor_picker_hexa",
                                    DEFAULT_PICKER_HEXA,
                                ),
                                update_modelValue=self._picker_update_js(),
                                mode="hexa",
                                hide_inputs=False,
                                show_swatches=True,
                                elevation=0,
                                width=320,
                                canvas_height=180,
                            )
                        with v3.VCardActions(classes="pa-2 pt-0"):
                            v3.VSpacer()
                            v3.VBtn(
                                "Done",
                                variant="text",
                                click="style_editor_picker_open = false",
                            )
                html.Div(
                    "{{ style_editor_error }}",
                    v_if="style_editor_error",
                    classes="text-error text-caption mt-2",
                )
                with html.Div(classes="d-flex justify-end mt-3"):
                    v3.VBtn(
                        "Apply Styles",
                        color="primary",
                        click=(self.apply_all, "[style_editor_controls]"),
                        disabled=("style_editor_controls.length === 0",),
                        size="large",
                    )

    def open(self):
        self.refresh()
        self._state.show_style_editor = True

    def refresh(self):
        with self._state:
            self._state.style_editor_controls = self._build_controls(
                self._state.field_specs
            )
            self._state.style_editor_error = ""

    def apply_all(self, controls):
        controls = controls or []
        updated_specs = copy.deepcopy(self._state.field_specs)

        for field_control in controls:
            field_name = field_control.get("name", "")
            if field_name not in updated_specs:
                self._state.style_editor_error = f"Unknown field '{field_name}'."
                return
            styles = copy.deepcopy(updated_specs[field_name].get("styles") or {})

            if not self._apply_dataset_groups(
                styles,
                field_name,
                field_control.get("color_groups", []),
                self._parse_color_entry,
            ):
                return
            if not self._apply_dataset_groups(
                styles,
                field_name,
                field_control.get("width_groups", []),
                self._parse_width_entry,
            ):
                return
            if not self._apply_colormap_groups(styles, field_control, field_name):
                return

            updated_specs[field_name]["styles"] = styles

        with self._state:
            self._state.field_specs = updated_specs
            self._state.style_editor_error = ""

    def _build_controls(self, specs):
        controls = []
        columns = self._dataset_columns()
        for field_name in sorted(
            specs.keys(),
            key=lambda name: specs.get(name, {}).get("priority", 999),
        ):
            spec = specs.get(field_name, {})
            styles = spec.get("styles") or {}
            color_groups = self._build_color_groups(styles, columns)
            width_groups = self._build_width_groups(styles, columns)
            colormap_groups = self._build_colormap_groups(styles)

            controls.append(
                {
                    "name": field_name,
                    "label": spec.get("label", field_name),
                    "icon": spec.get("icon", "mdi-cog"),
                    "icon_color": styles.get("icon_color"),
                    "color_groups": color_groups,
                    "width_groups": width_groups,
                    "colormap_groups": colormap_groups,
                }
            )

        return controls

    def _dataset_columns(self):
        right = self._datasets[0]
        left = self._datasets[1]
        right_label = right.name if right.enabled and right.name else "Right"
        left_label = left.name if left.enabled and left.name else "Left"
        if left.enabled:
            # Match the main table ordering: left column first, right column second.
            return [(1, left_label), (0, right_label)]
        return [(0, right_label)]

    def _apply_dataset_groups(self, styles, field_name, groups, value_parser):
        for group in groups:
            key = group.get("key")
            if key not in styles:
                continue
            values = normalize_dataset_values(styles.get(key), styles.get(key))
            for entry in group.get("values", []):
                idx = self._entry_index(entry)
                if idx is None:
                    continue
                try:
                    values[idx] = value_parser(entry)
                except ValueError as exc:
                    self._state.style_editor_error = f"{field_name}.{key}[{idx}] {exc}"
                    return False
            styles[key] = values
        return True

    def _apply_colormap_groups(self, styles, field_control, field_name):
        for group in field_control.get("colormap_groups", []):
            key = group.get("key")
            if key not in styles:
                continue
            try:
                value_min = float(group.get("min"))
                value_max = float(group.get("max"))
            except (TypeError, ValueError):
                self._state.style_editor_error = (
                    f"{field_name}.{key} min/max must be numeric."
                )
                return False
            if value_min >= value_max:
                self._state.style_editor_error = (
                    f"{field_name}.{key} min must be less than max."
                )
                return False
            current = styles.get(key)
            updated = dict(current) if isinstance(current, dict) else {}
            if "palette" not in updated:
                updated["palette"] = "RDBU_11"
            updated["min"] = value_min
            updated["max"] = value_max
            styles[key] = updated
        return True

    def _build_color_groups(self, styles, columns):
        groups = []
        for key, label in COLOR_STYLE_GROUPS:
            group = self._build_dataset_group(
                styles,
                key=key,
                label=label,
                columns=columns,
                value_key="hexa",
                formatter=self._rgba_to_hexa,
            )
            if group is not None:
                groups.append(group)
        return groups

    def _build_width_groups(self, styles, columns):
        group = self._build_dataset_group(
            styles,
            key="line_width",
            label="Line width",
            columns=columns,
            value_key="value",
            formatter=self._to_float,
        )
        return [] if group is None else [group]

    def _build_colormap_groups(self, styles):
        color_map_range = styles.get("color_map_range")
        if not isinstance(color_map_range, dict):
            return []
        try:
            value_min = float(color_map_range.get("min"))
            value_max = float(color_map_range.get("max"))
        except (TypeError, ValueError):
            return []
        palette_name = str(color_map_range.get("palette", "RDBU_11"))
        return [
            {
                "key": "color_map_range",
                "label": "Color map range",
                "gradient": self._palette_gradient(palette_name),
                "min": value_min,
                "max": value_max,
            }
        ]

    def _build_dataset_group(self, styles, key, label, columns, value_key, formatter):
        value = styles.get(key)
        if not isinstance(value, list | tuple):
            return None
        pair = normalize_dataset_values(value, value)
        return {
            "key": key,
            "label": label,
            "values": self._build_dataset_entries(pair, columns, value_key, formatter),
        }

    @staticmethod
    def _build_dataset_entries(pair, columns, value_key, formatter):
        return [
            {
                "idx": idx,
                "label": column_label,
                value_key: formatter(pair[idx]),
            }
            for idx, column_label in columns
        ]

    @staticmethod
    def _entry_index(entry):
        try:
            idx = int(entry.get("idx", -1))
        except (TypeError, ValueError):
            return None
        return idx if idx in (0, 1) else None

    @staticmethod
    def _open_picker_js():
        return (
            "style_editor_picker_open = true;"
            "style_editor_picker_field_idx = field_idx;"
            "style_editor_picker_group_idx = group_idx;"
            "style_editor_picker_entry_idx = entry_idx;"
            "style_editor_picker_hexa = entry.hexa;"
        )

    @staticmethod
    def _picker_update_js():
        return """
(value) => {{
  style_editor_picker_hexa = value;
  const field = style_editor_controls?.[style_editor_picker_field_idx];
  const group = field?.color_groups?.[style_editor_picker_group_idx];
  const entry = group?.values?.[style_editor_picker_entry_idx];
  if (entry) {{
    entry.hexa = value;
  }}
}}
        """

    def _parse_color_entry(self, entry):
        return self._parse_color(entry.get("hexa", "#000000ff"))

    @staticmethod
    def _parse_width_entry(entry):
        try:
            return float(entry.get("value", "0"))
        except (TypeError, ValueError) as exc:
            msg = "must be numeric."
            raise ValueError(msg) from exc

    @staticmethod
    def _to_float(value):
        try:
            return float(value)
        except (TypeError, ValueError):
            return 0.0

    @staticmethod
    def _rgba_to_hexa(value):
        rgba = list(value) if isinstance(value, list | tuple) else [0, 0, 0, 255]
        while len(rgba) < 4:
            rgba.append(255)
        r, g, b, a = (max(0, min(255, int(float(c)))) for c in rgba[:4])
        return f"#{r:02x}{g:02x}{b:02x}{a:02x}"

    @staticmethod
    def _parse_color(hexa_value):
        value = str(hexa_value or "").strip()
        if value.startswith("#"):
            value = value[1:]
        if len(value) == 6:
            value += "ff"
        if len(value) != 8:
            msg = "color must be #RRGGBBAA (or #RRGGBB)."
            raise ValueError(msg)
        try:
            r = int(value[0:2], 16)
            g = int(value[2:4], 16)
            b = int(value[4:6], 16)
            a = int(value[6:8], 16)
        except ValueError as exc:
            msg = "invalid color channel."
            raise ValueError(msg) from exc
        return [r, g, b, a]

    @staticmethod
    def _palette_gradient(palette_name):
        palette = PALETTES.get(palette_name, RDBU_11)
        if not palette:
            return "linear-gradient(90deg, #000000, #ffffff)"
        if len(palette) == 1:
            r, g, b = palette[0]
            return f"rgb({r}, {g}, {b})"
        stop_count = len(palette) - 1
        stops = []
        for idx, (r, g, b) in enumerate(palette):
            pct = round((idx / stop_count) * 100)
            stops.append(f"rgb({r}, {g}, {b}) {pct}%")
        return f"linear-gradient(90deg, {', '.join(stops)})"



import numpy as np
import pandas as pd

from fennil.app.deck.primitives import line_layers, polygon_layers

from .styles import (
    FAULT_PROJ_LINE_WIDTH,
    SLIP_NEGATIVE_COLOR,
    SLIP_NEGATIVE_EXTREME_COLOR,
    SLIP_POSITIVE_COLOR,
    SLIP_POSITIVE_EXTREME_COLOR,
    SLIP_WIDTH_CAP_MM_PER_YR,
    SLIP_WIDTH_MIN_PIXELS,
    SLIP_WIDTH_SCALE,
)
from .tooltips import format_segment_tooltip

REQUIRED_SEG_COLS = {
    "model_strike_slip_rate",
    "model_dip_slip_rate",
    "model_tensile_slip_rate",
}


def fault_line_dataframe(segment, seg_tooltip_enabled):
    fault_lines_df = pd.DataFrame(
        {
            "start_lon": segment.lon1.to_numpy(),
            "start_lat": segment.lat1.to_numpy(),
            "end_lon": segment.lon2.to_numpy(),
            "end_lat": segment.lat2.to_numpy(),
        }
    )

    if seg_tooltip_enabled and REQUIRED_SEG_COLS.issubset(segment.columns):
        fault_lines_df["tooltip"] = [
            format_segment_tooltip(
                name,
                lon1,
                lat1,
                lon2,
                lat2,
                ss_rate,
                ds_rate,
                ts_rate,
            )
            for name, lon1, lat1, lon2, lat2, ss_rate, ds_rate, ts_rate in zip(
                segment.name.to_numpy(),
                segment.lon1.to_numpy(),
                segment.lat1.to_numpy(),
                segment.lon2.to_numpy(),
                segment.lat2.to_numpy(),
                segment.model_strike_slip_rate.to_numpy(),
                segment.model_dip_slip_rate.to_numpy(),
                segment.model_tensile_slip_rate.to_numpy(),
                strict=False,
            )
        ]

    return fault_lines_df


def fault_line_layers(folder_number, segment, seg_tooltip_enabled, color, line_width):
    fault_lines_df = fault_line_dataframe(segment, seg_tooltip_enabled)

    layers = line_layers(
        "fault",
        fault_lines_df,
        color,
        line_width,
        folder_number,
        width_min_pixels=1,
        pickable=seg_tooltip_enabled,
    )
    return layers, fault_lines_df


def segment_slip_layers(
    folder_number,
    segment,
    seg_slip_type,
    seg_tooltip_enabled,
    fault_lines_df,
    velocity_scale=1.0,
):
    velocity_scale = 1.0 if velocity_scale is None else float(velocity_scale)

    if seg_slip_type == "ss":
        slip_values = segment.model_strike_slip_rate.to_numpy()
    else:
        slip_values = (
            segment.model_dip_slip_rate.to_numpy()
            - segment.model_tensile_slip_rate.to_numpy()
        )

    slip_values = np.asarray(slip_values)
    slip_values = np.nan_to_num(slip_values, nan=0.0, posinf=0.0, neginf=0.0)

    seg_lines_df = pd.DataFrame(
        {
            "start_lon": segment.lon1.to_numpy(),
            "start_lat": segment.lat1.to_numpy(),
            "end_lon": segment.lon2.to_numpy(),
            "end_lat": segment.lat2.to_numpy(),
            "slip_rate": slip_values,
            "line_width": np.clip(np.abs(slip_values), 0.0, SLIP_WIDTH_CAP_MM_PER_YR),
        }
    )

    def _slip_color(value):
        if value < -SLIP_WIDTH_CAP_MM_PER_YR:
            return SLIP_NEGATIVE_EXTREME_COLOR
        if value > SLIP_WIDTH_CAP_MM_PER_YR:
            return SLIP_POSITIVE_EXTREME_COLOR
        if value < 0:
            return SLIP_NEGATIVE_COLOR
        return SLIP_POSITIVE_COLOR

    seg_lines_df["color"] = [_slip_color(value) for value in slip_values]
    if seg_tooltip_enabled and "tooltip" in fault_lines_df.columns:
        seg_lines_df["tooltip"] = fault_lines_df["tooltip"].to_numpy()

    return line_layers(
        "segments",
        seg_lines_df,
        "color",
        "line_width",
        folder_number,
        width_min_pixels=SLIP_WIDTH_MIN_PIXELS,
        width_scale=SLIP_WIDTH_SCALE * velocity_scale,
        width_units="pixels",
        pickable=seg_tooltip_enabled,
    )


def fault_projection_layers(name, fault_proj_df, fill, line):
    if fault_proj_df is None or fault_proj_df.empty:
        return []
    return polygon_layers(
        "fault_proj",
        fault_proj_df,
        fill,
        line,
        FAULT_PROJ_LINE_WIDTH,
        name,
        pickable=False,
    )



from pathlib import Path

from trame.widgets import dataclass, html
from trame.widgets import vuetify3 as v3
from trame_dataclass.core import StateDataModel

from fennil.app.io import is_valid_data_folder

FILE_BROWSER_HEADERS = [
    {"title": "Name", "align": "start", "key": "name", "sortable": False},
    {"title": "Type", "align": "start", "key": "type", "sortable": False},
]


class FileBrowserState(StateDataModel):
    show: bool = False
    current: str = "/"
    listing: list
    active: int = -1
    error: str | None
    headers: list = FILE_BROWSER_HEADERS


class FileBrowser(dataclass.Provider):
    def __init__(self, current_directory=None, on_open=None, **kwargs):
        if current_directory is None:
            current_directory = Path.cwd()
        self._on_open = on_open
        self._state = None
        super().__init__(name="browser", **kwargs)
        self._state = FileBrowserState(self.server, current=str(current_directory))
        self.instance = self._state._id
        self.update_listing()

        with (
            self,
            v3.VDialog(
                v_model="browser.show",
                max_width="900",
                persistent=True,
            ),
        ):
            with v3.VCard(title="Select data folder", rounded="lg"):
                with v3.VCardText():
                    with v3.VRow(dense=True, classes="pb-1 align-center"):
                        v3.VBtn(
                            icon="mdi-home",
                            variant="text",
                            size="small",
                            click=self.go_home,
                        )
                        v3.VBtn(
                            icon="mdi-folder-upload-outline",
                            variant="text",
                            size="small",
                            click=self.go_parent,
                        )
                        v3.VTextField(
                            v_model="browser.current",
                            hide_details=True,
                            density="compact",
                            variant="outlined",
                            readonly=True,
                            classes="ml-2 flex-grow-1",
                        )
                    with v3.VDataTable(
                        density="compact",
                        fixed_header=True,
                        headers=["browser.header"],
                        items=["browser.listing"],
                        height="50vh",
                        style="user-select: none; cursor: pointer;",
                        items_per_page=-1,
                    ):
                        v3.Template(v_slot_bottom=True)
                        with v3.Template(v_slot_item="{ item }"):
                            with v3.VDataTableRow(
                                item=["item"],
                                click=(self.select_entry, "[item]"),
                                dblclick=(self.open_entry, "[item]"),
                                classes=[
                                    "{ 'bg-grey-lighten-3': item.index === browser.active }"
                                ],
                            ):
                                with v3.Template(raw_attrs=["v-slot:item.name"]):
                                    with html.Div(classes="d-flex align-center"):
                                        v3.VIcon(
                                            "{{ item.icon }}",
                                            size="small",
                                            classes="mr-2",
                                        )
                                        html.Div("{{ item.name }}")
                                with v3.Template(raw_attrs=["v-slot:item.type"]):
                                    html.Div("{{ item.type }}")

                with v3.VCardActions(classes="pa-3"):
                    html.Div(
                        "{{ browser.error }}",
                        v_if="browser.error",
                        classes="text-error text-caption",
                    )
                    v3.VSpacer()
                    v3.VBtn(
                        text="Cancel",
                        variant="flat",
                        click="browser.show = false",
                    )
                    v3.VBtn(
                        text="Select folder",
                        color="primary",
                        variant="flat",
                        click=self.select_folder,
                    )

    def update_listing(self):
        current = Path(self._state.current)
        if not current.exists():
            current = Path.home()
            self._state.current = str(current.resolve())

        entries = []
        for entry in current.iterdir():
            name = entry.name
            if name.startswith("."):
                continue
            if entry.is_dir():
                entries.append(
                    {
                        "name": name,
                        "type": "directory",
                        "icon": "mdi-folder",
                    }
                )
            elif entry.is_file():
                entries.append(
                    {
                        "name": name,
                        "type": "file",
                        "icon": "mdi-file-document-outline",
                    }
                )
        entries.sort(key=lambda item: (item["type"] != "directory", item["name"]))
        listing = [{**item, "index": idx} for idx, item in enumerate(entries)]
        self._state.listing = listing
        self._state.active = -1

    def select_entry(self, entry):
        self._state.active = entry.get("index", -1) if entry else -1

    def open_entry(self, entry):
        if not entry or entry.get("type") != "directory":
            return
        current = Path(self._state.current)
        next_path = (current / entry.get("name")).resolve()

        if is_valid_data_folder(next_path):
            self._state.error = None
            self._state.show = False
            if self._on_open:
                self._on_open(next_path)
            return

        self._state.current = str(next_path)
        self.update_listing()

    def go_home(self):
        self._state.current = str(Path.home().resolve())
        self.update_listing()

    def go_parent(self):
        current = Path(self._state.current)
        parent = current.parent if current.parent != current else current
        self._state.current = str(parent.resolve())
        self.update_listing()

    def open(self, existing_path=None):
        if existing_path:
            self._state.current = str(Path(existing_path).resolve())

        self._state.show = True
        self._state.error = None
        self.update_listing()

    def select_folder(self):
        current = Path(self._state.current)
        folder_path = current
        active_idx = self._state.active
        listing = self._state.listing
        if active_idx is not None and active_idx >= 0:
            if active_idx >= len(listing):
                active_idx = -1
            else:
                entry = listing[active_idx]
                if entry.get("type") == "directory":
                    folder_path = (current / entry.get("name")).resolve()
        if not is_valid_data_folder(folder_path):
            self._state.error = "Selected folder is missing required model_*.csv files."
            return
        self._state.error = None
        self._state.show = False
        if self._on_open:
            self._on_open(folder_path)



from fennil.app.deck.primitives import line_layers, polygon_layers

from .styles import BLACK, RED, map_coupling_colors, map_slip_colors


def tde_mesh_layers(folder_number, tde_df, slip_values, value_min, value_max):
    if tde_df is None or tde_df.empty:
        return []
    tde_df = tde_df.copy()
    tde_df["color"] = map_slip_colors(
        slip_values,
        value_min=value_min,
        value_max=value_max,
    )
    return polygon_layers(
        "tde",
        tde_df,
        "color",
        [0, 0, 0, 0],
        0,
        folder_number,
        line_width_min_pixels=0,
        stroked=False,
        pickable=False,
    )


def tde_coupling_layers(folder_number, tde_df, coupling_values, value_min, value_max):
    if tde_df is None or tde_df.empty:
        return []
    tde_df = tde_df.copy()
    tde_df["color"] = map_coupling_colors(
        coupling_values,
        value_min=value_min,
        value_max=value_max,
    )
    return polygon_layers(
        "tde_coupling",
        tde_df,
        "color",
        [0, 0, 0, 0],
        0,
        folder_number,
        line_width_min_pixels=0,
        stroked=False,
        pickable=False,
    )


def tde_perimeter_layers(folder_number, tde_perim_df):
    if tde_perim_df is None or tde_perim_df.empty:
        return []
    tde_perim_df = tde_perim_df.copy()
    tde_perim_df["color"] = [
        RED if int(flag) == 1 else BLACK for flag in tde_perim_df["proj_col"]
    ]
    return line_layers(
        "tde_perim",
        tde_perim_df,
        "color",
        1,
        folder_number,
        width_min_pixels=1,
        pickable=False,
    )



from urllib.parse import quote

import numpy as np
import pandas as pd

from fennil.app.deck.primitives import icon_layers

from .styles import (
    RDBU_11,
    RES_COMPARE_DIFF_MAX,
    RES_COMPARE_DIFF_MIN,
    RES_COMPARE_SIZE_SCALE,
    RES_COMPARE_UNIQUE_COLOR,
    RES_COMPARE_UNIQUE_SIZE_PIXELS,
)

CIRCLE_ICON = {
    "url": "data:image/svg+xml;utf8,"
    + quote(
        "<svg xmlns='http://www.w3.org/2000/svg' width='64' height='64' viewBox='0 0 64 64'>"
        "<circle cx='32' cy='32' r='28' fill='black'/>"
        "</svg>"
    ),
    "width": 64,
    "height": 64,
    "anchorX": 32,
    "anchorY": 32,
    "mask": True,
}


def _map_residual_diff_colors(values, value_min, value_max):
    colors = []
    span = value_max - value_min
    if span <= 0:
        span = 1.0
    for raw_value in values:
        value = raw_value if np.isfinite(raw_value) else 0.0
        value = float(np.clip(value, value_min, value_max))
        position = (value - value_min) / span
        color_index = int(np.floor(position * len(RDBU_11)))
        color_index = max(0, min(len(RDBU_11) - 1, color_index))
        r, g, b = RDBU_11[color_index]
        colors.append([r, g, b, 220])
    return colors


def _residual_station_data(dataset):
    return pd.DataFrame(
        {
            "lon": dataset.station.lon.to_numpy(),
            "lat": dataset.station.lat.to_numpy(),
            "res_mag": dataset.resmag,
        }
    )


def residual_compare_layers(
    right_dataset,
    left_dataset,
    velocity_scale,
    value_min=RES_COMPARE_DIFF_MIN,
    value_max=RES_COMPARE_DIFF_MAX,
):
    right = _residual_station_data(right_dataset)
    left = _residual_station_data(left_dataset)

    common = right.merge(
        left,
        how="inner",
        on=["lon", "lat"],
        suffixes=("_1", "_2"),
    )
    unique = (
        pd.concat(
            (
                right[["lon", "lat"]],
                left[["lon", "lat"]],
            ),
            ignore_index=True,
        )
        .drop_duplicates(keep=False, ignore_index=True)
        .reset_index(drop=True)
    )

    layers = []

    if not common.empty:
        res_mag_diff = common["res_mag_2"].to_numpy() - common["res_mag_1"].to_numpy()
        sized_res_mag_diff = (
            np.abs(res_mag_diff) * RES_COMPARE_SIZE_SCALE * velocity_scale
        )

        common_df = pd.DataFrame(
            {
                "lon": common["lon"].to_numpy(),
                "lat": common["lat"].to_numpy(),
                "res_mag_diff": res_mag_diff,
                "size": sized_res_mag_diff,
                "color": _map_residual_diff_colors(res_mag_diff, value_min, value_max),
                "icon": [CIRCLE_ICON] * len(common),
            }
        )
        common_df["tooltip"] = [
            f"<b>Resid. diff</b>: {value:.4f} mm/yr" for value in res_mag_diff
        ]

        layers.extend(
            icon_layers(
                "res_compare_common",
                common_df,
                get_position=["lon", "lat"],
                get_icon="icon",
                get_color="color",
                get_size="size",
                folder_number="compare",
                position_lon_key="lon",
                size_min_pixels=0,
                pickable=True,
            )
        )

    if not unique.empty:
        unique_df = unique.copy()
        unique_df["icon"] = [CIRCLE_ICON] * len(unique_df)
        unique_df["tooltip"] = [
            "<b>Unique station</b>: present in only one dataset"
        ] * len(unique_df)

        layers.extend(
            icon_layers(
                "res_compare_unique",
                unique_df,
                get_position=["lon", "lat"],
                get_icon="icon",
                get_color=RES_COMPARE_UNIQUE_COLOR,
                get_size=RES_COMPARE_UNIQUE_SIZE_PIXELS,
                folder_number="compare",
                position_lon_key="lon",
                pickable=True,
            )
        )

    return layers



from fennil.app.io import Dataset
from fennil.app.registry import FieldSpec, LayerContext
from fennil.app.viz.slip_compare import REQUIRED_SLIP_COMPARE_COLS, slip_compare_layers

SPEC = FieldSpec(
    priority=51,
    label="Slip compare",
    icon="mdi-align-vertical-center",
    ui_type="VBtnToggle",
    options=[
        {"text": "SS", "value": "ss"},
        {"text": "DS", "value": "ds"},
    ],
    default=None,
    styles={
        "icon_color": "rgba(44, 160, 44, 0.78)",
    },
    multiple=False,
)


def builder(name: str, ctx: LayerContext):
    right = ctx.datasets[0]
    left = ctx.datasets[1]
    if not (right.enabled and left.enabled):
        return
    if right.data is None or left.data is None:
        return

    slip_type = left.fields.get(name)
    if slip_type not in {"ss", "ds"}:
        return

    ctx.layers.extend(
        slip_compare_layers(
            right.data,
            left.data,
            slip_type,
            ctx.velocity_scale,
        )
    )


def can_render(dataset: Dataset) -> bool:
    return dataset is not None and REQUIRED_SLIP_COMPARE_COLS.issubset(
        dataset.segment.columns
    )



import numpy as np

VELOCITY_SCALE = 1000
VECTOR_ARROW_SIZE_FACTOR = 6
VECTOR_ARROW_MIN_PIXELS = 8
VECTOR_ARROW_MAX_PIXELS = 28

RES_COMPARE_SIZE_SCALE = VELOCITY_SCALE / 400.0
RES_COMPARE_DIFF_MIN = -5.0
RES_COMPARE_DIFF_MAX = 5.0
RES_COMPARE_UNIQUE_SIZE_PIXELS = 15.0
RES_COMPARE_UNIQUE_COLOR = [0, 0, 0, 220]

FAULT_PROJ_LINE_WIDTH = 1
SLIP_WIDTH_SCALE = 0.05
SLIP_WIDTH_MIN_PIXELS = 1
SLIP_WIDTH_CAP_MM_PER_YR = 400.0
SLIP_NEGATIVE_COLOR = [31, 119, 180, 220]  # tab:blue
SLIP_POSITIVE_COLOR = [255, 127, 14, 220]  # tab:orange
SLIP_NEGATIVE_EXTREME_COLOR = [25, 230, 255, 220]  # bright blue
SLIP_POSITIVE_EXTREME_COLOR = [255, 40, 40, 220]  # bright red

SLIP_COMPARE_MATCH_TOL_DEG = 1.0e-4
SLIP_COMPARE_WIDTH_SCALE = SLIP_WIDTH_SCALE * 128
SLIP_COMPARE_WIDTH_MIN_PIXELS = 1
SLIP_COMPARE_FASTER_COLOR = [44, 160, 44, 220]  # green
SLIP_COMPARE_SLOWER_COLOR = [214, 39, 40, 220]  # red
SLIP_COMPARE_NEUTRAL_COLOR = [140, 140, 140, 220]


# ColorBrewer RdBu[11] palette for discrete slip-rate coloring
RDBU_11 = [
    (5, 48, 97),
    (33, 102, 172),
    (67, 147, 195),
    (146, 197, 222),
    (209, 229, 240),
    (247, 247, 247),
    (253, 219, 199),
    (244, 165, 130),
    (214, 96, 77),
    (178, 24, 43),
    (103, 0, 31),
]
RED = [255, 0, 0, 255]
BLACK = [0, 0, 0, 255]

SLIP_RATE_MIN = -100.0
SLIP_RATE_MAX = 100.0
COUPLING_MIN = -1.0
COUPLING_MAX = 1.0


def map_discrete_colors(values, value_min, value_max, palette):
    colors_array = []
    span = value_max - value_min
    if span <= 0:
        span = 1.0
    for raw_value in values:
        value = raw_value
        if not np.isfinite(value):
            value = 0.0
        value = float(np.clip(value, value_min, value_max))
        position = (value - value_min) / span
        index = int(np.floor(position * len(palette)))
        index = max(0, min(len(palette) - 1, index))
        r, g, b = palette[index]
        colors_array.append([r, g, b, 255])
    return colors_array


def map_slip_colors(values, value_min=SLIP_RATE_MIN, value_max=SLIP_RATE_MAX):
    return map_discrete_colors(values, value_min, value_max, RDBU_11)


def map_coupling_colors(values, value_min=COUPLING_MIN, value_max=COUPLING_MAX):
    return map_discrete_colors(values, value_min, value_max, RDBU_11)


def normalize_dataset_values(raw_values, default_values):
    """Return exactly two style values ordered as [right, left]."""
    values = list(raw_values) if isinstance(raw_values, list | tuple) else []

    if not values:
        values = list(default_values)

    if not values:
        return [None, None]

    if len(values) == 1:
        return [values[0], values[0]]

    return [values[0], values[1]]


def dataset_value(two_values, dataset_index):
    return two_values[0] if dataset_index == 0 else two_values[1]


def dataset_float_value(two_values, dataset_index, fallback):
    raw_value = dataset_value(two_values, dataset_index)
    try:
        return float(raw_value)
    except (TypeError, ValueError):
        return float(fallback)



import numpy as np
import pandas as pd

from fennil.app.deck.primitives import line_layers

from .styles import (
    FAULT_PROJ_LINE_WIDTH,
    SLIP_COMPARE_FASTER_COLOR,
    SLIP_COMPARE_MATCH_TOL_DEG,
    SLIP_COMPARE_NEUTRAL_COLOR,
    SLIP_COMPARE_SLOWER_COLOR,
    SLIP_COMPARE_WIDTH_MIN_PIXELS,
    SLIP_COMPARE_WIDTH_SCALE,
)

REQUIRED_SLIP_COMPARE_COLS = {
    "lon1",
    "lat1",
    "lon2",
    "lat2",
    "model_strike_slip_rate",
    "model_dip_slip_rate",
    "model_tensile_slip_rate",
}


def slip_compare_layers(right_dataset, left_dataset, slip_type, velocity_scale):
    velocity_scale = 1.0 if velocity_scale is None else float(velocity_scale)

    right_df = _segment_slip_frame(right_dataset.segment, slip_type)
    left_df = _segment_slip_frame(left_dataset.segment, slip_type)

    shared = right_df.merge(left_df, on="segment_key", suffixes=("_right", "_left"))
    shared_keys = set(shared["segment_key"])

    layers = []

    if not shared.empty:
        diff = (
            shared["slip_rate_left"].to_numpy() - shared["slip_rate_right"].to_numpy()
        )
        shared_df = pd.DataFrame(
            {
                "start_lon": shared["start_lon_right"].to_numpy(),
                "start_lat": shared["start_lat_right"].to_numpy(),
                "end_lon": shared["end_lon_right"].to_numpy(),
                "end_lat": shared["end_lat_right"].to_numpy(),
                "line_width": np.abs(diff),
                "color": [_diff_color(value) for value in diff],
            }
        )
        shared_df["tooltip"] = [
            _shared_tooltip(name_left, name_right, left_rate, right_rate, diff_rate)
            for name_left, name_right, left_rate, right_rate, diff_rate in zip(
                shared["name_left"].to_numpy(),
                shared["name_right"].to_numpy(),
                shared["slip_rate_left"].to_numpy(),
                shared["slip_rate_right"].to_numpy(),
                diff,
                strict=False,
            )
        ]
        layers.extend(
            line_layers(
                "slip_compare_shared",
                shared_df,
                "color",
                "line_width",
                "compare",
                width_min_pixels=SLIP_COMPARE_WIDTH_MIN_PIXELS,
                width_scale=SLIP_COMPARE_WIDTH_SCALE * velocity_scale,
                width_units="pixels",
                pickable=True,
            )
        )

    unmatched_right = _unmatched_segments_df(right_df, shared_keys, "right")
    unmatched_left = _unmatched_segments_df(left_df, shared_keys, "left")
    unmatched_df = pd.concat((unmatched_right, unmatched_left), ignore_index=True)
    if not unmatched_df.empty:
        layers.extend(
            line_layers(
                "slip_compare_unmatched",
                unmatched_df,
                "color",
                "line_width",
                "compare",
                width_min_pixels=FAULT_PROJ_LINE_WIDTH,
                width_scale=1,
                width_units="pixels",
                pickable=True,
            )
        )

    return layers


def _segment_slip_frame(segment, slip_type):
    frame = _build_segment_rows(segment, slip_type)
    frame = _finite_segment_rows(frame)
    if frame.empty:
        frame["segment_key"] = pd.Series(dtype=str)
        return frame

    frame = _with_unordered_segment_keys(frame)
    return frame.groupby("segment_key", as_index=False).agg(
        start_lon=("start_lon", "first"),
        start_lat=("start_lat", "first"),
        end_lon=("end_lon", "first"),
        end_lat=("end_lat", "first"),
        slip_rate=("slip_rate", "mean"),
        name=("name", "first"),
    )


def _build_segment_rows(segment, slip_type):
    return pd.DataFrame(
        {
            "start_lon": segment["lon1"].to_numpy(dtype=float),
            "start_lat": segment["lat1"].to_numpy(dtype=float),
            "end_lon": segment["lon2"].to_numpy(dtype=float),
            "end_lat": segment["lat2"].to_numpy(dtype=float),
            "slip_rate": _segment_slip_values(segment, slip_type),
            "name": (
                segment["name"].fillna("").astype(str).to_numpy()
                if "name" in segment.columns
                else np.array([""] * len(segment), dtype=object)
            ),
        }
    )


def _finite_segment_rows(frame):
    valid_mask = (
        np.isfinite(frame["start_lon"].to_numpy())
        & np.isfinite(frame["start_lat"].to_numpy())
        & np.isfinite(frame["end_lon"].to_numpy())
        & np.isfinite(frame["end_lat"].to_numpy())
        & np.isfinite(frame["slip_rate"].to_numpy())
    )
    return frame.loc[valid_mask].copy()


def _with_unordered_segment_keys(frame):
    start_lon_q = _quantized(frame["start_lon"].to_numpy())
    start_lat_q = _quantized(frame["start_lat"].to_numpy())
    end_lon_q = _quantized(frame["end_lon"].to_numpy())
    end_lat_q = _quantized(frame["end_lat"].to_numpy())

    endpoint_a = _endpoint_key(start_lon_q, start_lat_q)
    endpoint_b = _endpoint_key(end_lon_q, end_lat_q)
    segment_keys = [
        _unordered_pair_key(a, b) for a, b in zip(endpoint_a, endpoint_b, strict=False)
    ]

    keyed_rows = frame.copy()
    keyed_rows["segment_key"] = segment_keys
    return keyed_rows


def _quantized(values):
    return np.rint(values / SLIP_COMPARE_MATCH_TOL_DEG)


def _endpoint_key(lon_q, lat_q):
    return lon_q.astype(np.int64).astype(str) + ":" + lat_q.astype(np.int64).astype(str)


def _unordered_pair_key(endpoint_a, endpoint_b):
    if endpoint_a <= endpoint_b:
        return f"{endpoint_a}:{endpoint_b}"
    return f"{endpoint_b}:{endpoint_a}"


def _unmatched_segments_df(source_df, shared_keys, model_name):
    if source_df.empty:
        return pd.DataFrame()

    unmatched = source_df[~source_df["segment_key"].isin(shared_keys)].copy()
    if unmatched.empty:
        return pd.DataFrame()

    unmatched["line_width"] = float(FAULT_PROJ_LINE_WIDTH)
    unmatched["color"] = [SLIP_COMPARE_NEUTRAL_COLOR] * len(unmatched)
    unmatched["tooltip"] = [
        _unmatched_tooltip(model_name, name, slip)
        for name, slip in zip(
            unmatched["name"].to_numpy(),
            unmatched["slip_rate"].to_numpy(),
            strict=False,
        )
    ]
    return unmatched[
        [
            "start_lon",
            "start_lat",
            "end_lon",
            "end_lat",
            "line_width",
            "color",
            "tooltip",
        ]
    ].reset_index(drop=True)


def _segment_slip_values(segment, slip_type):
    if slip_type == "ss":
        values = segment["model_strike_slip_rate"].to_numpy(dtype=float)
    else:
        values = segment["model_dip_slip_rate"].to_numpy(dtype=float) - segment[
            "model_tensile_slip_rate"
        ].to_numpy(dtype=float)
    return np.nan_to_num(values, nan=0.0, posinf=0.0, neginf=0.0)


def _diff_color(diff_rate):
    if diff_rate > 0:
        return SLIP_COMPARE_FASTER_COLOR
    if diff_rate < 0:
        return SLIP_COMPARE_SLOWER_COLOR
    return SLIP_COMPARE_NEUTRAL_COLOR


def _shared_tooltip(name_left, name_right, left_rate, right_rate, diff_rate):
    shared_name = name_left or name_right or "n/a"
    return (
        f"<b>Name</b>: {shared_name}<br/>"
        f"<b>Left slip</b>: {left_rate:.4f} mm/yr<br/>"
        f"<b>Right slip</b>: {right_rate:.4f} mm/yr<br/>"
        f"<b>Diff (left-right)</b>: {diff_rate:.4f} mm/yr"
    )


def _unmatched_tooltip(model_name, name, slip):
    return (
        f"<b>Name</b>: {name or 'n/a'}<br/>"
        f"<b>Only in</b>: {model_name} model<br/>"
        f"<b>Slip</b>: {slip:.4f} mm/yr"
    )



from collections.abc import Callable
from dataclasses import dataclass
from typing import Any

from fennil.app.io import Dataset


@dataclass(frozen=True)
class FieldSpec:
    priority: int
    label: str
    icon: str
    ui_type: str
    options: list[dict[str, str]] | None
    default: bool | str | None
    styles: Any | None = None
    multiple: bool = True

    def to_dict(self):
        return {
            "priority": self.priority,
            "label": self.label,
            "icon": self.icon,
            "type": self.ui_type,
            "styles": self.styles,
            "options": self.options,
            "multiple": self.multiple,
        }


class LayerContext:
    def __init__(self, specs, datasets, velocity_scale):
        self.specs = specs
        self.datasets = datasets
        self.velocity_scale = velocity_scale
        self.tde_layers = []
        self.layers = []
        self.vector_layers = []

    @property
    def field_names(self):
        names = []
        seen = set()
        for dataset in self.datasets:
            if not dataset.enabled:
                continue
            for name in dataset.available_fields:
                if name in seen:
                    continue
                seen.add(name)
                names.append(name)
        return names

    @property
    def all_layers(self):
        return self.tde_layers + self.layers + self.vector_layers

    def skip(self, name):
        return all(not (ds.enabled and ds.fields.get(name)) for ds in self.datasets)

    def enabled_datasets(self, name):
        return (
            (i, ds)
            for i, ds in enumerate(self.datasets)
            if ds.enabled and ds.fields.get(name) and name in ds.available_fields
        )


class FieldRegistry:
    def __init__(self):
        self._specs: dict[str, FieldSpec] = {}
        self._builders: dict[str, Callable[[str, LayerContext], None]] = {}
        self._can_render: dict[str, Callable[[Dataset], bool]] = {}

    def register(self, field_name: str, spec: FieldSpec, builder, can_render):
        self._specs[field_name] = spec
        self._builders[field_name] = builder
        self._can_render[field_name] = can_render

    def field_defaults(self):
        return {name: spec.default for name, spec in self._specs.items()}

    def available_fields(self, dataset):
        name_priority = [
            (name, spec.priority)
            for name, spec in self._specs.items()
            if self._can_render[name](dataset)
        ]
        return [name for name, _ in sorted(name_priority, key=lambda tup: tup[1])]

    def export_specs(self):
        return {name: spec.to_dict() for name, spec in self._specs.items()}

    def build_layers(self, ctx: LayerContext):
        for name in ctx.field_names:
            builder = self._builders.get(name)
            if builder is None:
                continue
            builder(name, ctx)


FIELD_REGISTRY = FieldRegistry()



import pandas as pd

from fennil.app.deck.primitives import scatter_layers


def station_layers(folder_number, station, color):
    station_df = pd.DataFrame(
        {
            "lon": station.lon.to_numpy(),
            "lat": station.lat.to_numpy(),
            "name": station.name.to_numpy(),
        }
    )
    station_df["tooltip"] = [f"<b>Name</b>: {name}" for name in station_df["name"]]

    return scatter_layers(
        "stations",
        station_df,
        color,
        3000,
        folder_number,
        radius_min_pixels=2,
        radius_max_pixels=5,
        pickable=True,
    )
